"""This module contains the structure of the evaluator, the core of the FData
object for extrapolation and evaluation of FDataGrids"""

from abc import ABC, abstractmethod


class EvaluatorConstructor(ABC):
    """Constructor of an evaluator.

    A constructor builds an Evaluator from a :class:`FData`, which is
    used to the evaluation in the functional data object.

    The evaluator constructor should have a method :func:`evaluator` which
    receives an fdata object and returns an :class:`Evaluator`.

    """

    @abstractmethod
    def evaluator(self, fdata):
        """Construct an evaluator.

        Builds the evaluator from an functional data object.

        Args:
            fdata (:class:`FData`): Functional object where the evaluator will
                be used.

        Returns:
            (:class:`Evaluator`): Evaluator of the fdata.

        """
        pass

    def __eq__(self, other):
        """Equality operator between evaluators constructors"""
        return type(self) == type(other)


class Evaluator(ABC):
    """Structure of an evaluator.

    An evaluator defines how to evaluate points of a functional object, it
    can be used as extrapolator to evaluate points outside the domain range or
    as interpolator in a :class:`FDataGrid`. The corresponding examples of
    Interpolation and Extrapolation shows the basic usage of this class.

    The evaluator is called internally by :func:`evaluate`.

    Should implement the methods :func:`evaluate` and :func:`evaluate_composed`.


    """
    @abstractmethod
    def evaluate(self, eval_points, *, derivative=0):
        """Evaluation method.

        Evaluates the samples at the same evaluation points. The evaluation call
        will receive a 2-d array with the evaluation points.

        This method is called internally by :meth:`evaluate` when the argument
        `aligned_evaluation` is True.

        Args:
            eval_points (numpy.ndarray): Numpy array with shape
                `(len(eval_points), ndim_domain)` with the evaluation points.
                Each entry represents the coordinate of a point.
            derivative (int, optional): Order of the derivative. Defaults to 0.

        Returns:
            (numpy.darray): Numpy 3-d array with shape `(n_samples,
                len(eval_points), ndim_image)` with the result of the
                evaluation. The entry (i,j,k) will contain the value k-th
                image dimension of the i-th sample, at the j-th evaluation
                point.

        """
        pass

    @abstractmethod
    def evaluate_composed(self, eval_points, *, derivative=0):
        """Evaluation method.

        Evaluates the samples at different evaluation points. The evaluation
        call will receive a 3-d array with the evaluation points for each sample.

        This method is called internally by :meth:`evaluate` when the argument
        `aligned_evaluation` is False.

        Args:
            eval_points (numpy.ndarray): Numpy array with shape
                `(n_samples, number_eval_points, ndim_domain)` with the
                 evaluation points for each sample.
            derivative (int, optional): Order of the derivative. Defaults to 0.

        Returns:
            (numpy.darray): Numpy 3d array with shape `(n_samples,
                number_eval_points, ndim_image)` with the result of the
                evaluation. The entry (i,j,k) will contain the value k-th image
                dimension of the i-th sample, at the j-th evaluation point.

        """
        pass


class GenericEvaluator(Evaluator):
    """Generic Evaluator.

    Generic evaluator that recibes two functions to construct the evaluator.
    The functions will recieve an :class:`FData` as first argument, a numpy
    array with the eval_points and a named argument derivative.

    """

    def __init__(self, fdata, evaluate_func, evaluate_composed_func=None):
        self.fdata = fdata
        self.evaluate_func = evaluate_func

        if evaluate_composed_func is None:
            self.evaluate_composed_func = evaluate_func
        else:
            self.evaluate_composed_func = evaluate_composed_func

    def evaluate(self, eval_points, *, derivative=0):
        """Evaluation method.

        Evaluates the samples at the same evaluation points. The evaluation call
        will receive a 2-d array with the evaluation points.

        This method is called internally by :meth:`evaluate` when the argument
        `aligned_evaluation` is True.

        Args:
            eval_points (numpy.ndarray): Numpy array with shape
                `(len(eval_points), ndim_domain)` with the evaluation points.
                Each entry represents the coordinate of a point.
            derivative (int, optional): Order of the derivative. Defaults to 0.

        Returns:
            (numpy.darray): Numpy 3-d array with shape `(n_samples,
                len(eval_points), ndim_image)` with the result of the
                evaluation. The entry (i,j,k) will contain the value k-th
                image dimension of the i-th sample, at the j-th evaluation
                point.

        """
        return self.evaluate_func(self.fdata, eval_points,
                                  derivative=derivative)

    def evaluate_composed(self, eval_points, *, derivative=0):
        """Evaluation method.

        Evaluates the samples at different evaluation points. The evaluation
        call will receive a 3-d array with the evaluation points for each sample.

        This method is called internally by :meth:`evaluate` when the argument
        `aligned_evaluation` is False.

        Args:
            eval_points (numpy.ndarray): Numpy array with shape
                `(n_samples, number_eval_points, ndim_domain)` with the
                 evaluation points for each sample.
            derivative (int, optional): Order of the derivative. Defaults to 0.

        Returns:
            (numpy.darray): Numpy 3d array with shape `(n_samples,
                number_eval_points, ndim_image)` with the result of the
                evaluation. The entry (i,j,k) will contain the value k-th image
                dimension of the i-th sample, at the j-th evaluation point.

        """
        return self.evaluate_composed_func(self.fdata, eval_points,
                                           derivative=derivative)



"""Depth Measures Module.

This module includes different methods to order functional data,
from the center (larger values) outwards(smaller ones)."""

import numpy as np
from scipy.stats import rankdata
from .. import FDataGrid
import itertools
from functools import reduce

__author__ = "Amanda Hernando Bernabé"
__email__ = "amanda.hernando@estudiante.uam.es"


def _rank_samples(fdatagrid):
    """Ranks the he samples in the FDataGrid at each point of discretisation.

    Args:
        fdatagrid (FDataGrid): Object whose samples are ranked.

    Returns:
        numpy.darray: Array containing the ranks of the sample points.

    Examples:
        Univariate setting:

        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2], [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5], [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> sample_points = [0, 2, 4, 6, 8, 10]
        >>> fd = FDataGrid(data_matrix, sample_points)
        >>> _rank_samples(fd)
        array([[[ 4.],
                [ 4.],
                [ 4.],
                [ 4.],
                [ 4.],
                [ 4.]],
        <BLANKLINE>
               [[ 3.],
                [ 3.],
                [ 3.],
                [ 3.],
                [ 3.],
                [ 3.]],
        <BLANKLINE>
               [[ 1.],
                [ 1.],
                [ 2.],
                [ 2.],
                [ 2.],
                [ 2.]],
        <BLANKLINE>
               [[ 2.],
                [ 2.],
                [ 2.],
                [ 1.],
                [ 1.],
                [ 1.]]])

        Multivariate Setting:

        >>> data_matrix = [[[[1, 3], [2, 6]], [[23, 54], [43, 76]], [[2, 45], [12, 65]]],
        ...                [[[21, 34], [8, 16]], [[67, 43], [32, 21]], [[10, 24], [3, 12]]]]
        >>> sample_points = [[2, 4, 6], [3, 6]]
        >>> fd = FDataGrid(data_matrix, sample_points)
        >>> _rank_samples(fd)
        array([[[[ 1.,  1.],
                 [ 1.,  1.]],
        <BLANKLINE>
                [[ 1.,  2.],
                 [ 2.,  2.]],
        <BLANKLINE>
                [[ 1.,  2.],
                 [ 2.,  2.]]],
        <BLANKLINE>
        <BLANKLINE>
               [[[ 2.,  2.],
                 [ 2.,  2.]],
        <BLANKLINE>
                [[ 2.,  1.],
                 [ 1.,  1.]],
        <BLANKLINE>
                [[ 2.,  1.],
                 [ 1.,  1.]]]])
    """
    ranks = np.zeros(fdatagrid.shape)
    ncols_dim_image = np.asarray([range(fdatagrid.shape[i]) for i in range(len(fdatagrid.shape) - 1, 0, -1)])
    tuples = list(itertools.product(*ncols_dim_image))
    for t in tuples:
        ranks.T[t] = rankdata(fdatagrid.data_matrix.T[t], method='max')
    return ranks


def band_depth(fdatagrid, pointwise=False):
    """Implementation of Band Depth for functional data.

    The band depth of each sample is obtained by computing the fraction of the bands determined by two sample
    curves containing the whole graph of the first one. In the case the fdatagrid domain dimension is 2, instead
    of curves, surfaces determine the bands. In larger dimensions, the hyperplanes determine the bands.

    Args:
        fdatagrid (FDataGrid): Object over whose samples the band depth is going to be calculated.
        pointwise (boolean, optional): Indicates if the pointwise depth is also returned. Defaults to False.

    Returns:
        depth (numpy.darray): Array containing the band depth of the samples.

    Returns:
        depth_pointwise (numpy.darray, optional): Array containing the band depth of
        the samples at each point of discretisation. Only returned if pointwise equals to True.

    Examples:
        Univariate setting:

        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2], [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5], [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> sample_points = [0, 2, 4, 6, 8, 10]
        >>> fd = FDataGrid(data_matrix, sample_points)
        >>> band_depth(fd)
        array([[ 0.5       ],
               [ 0.83333333],
               [ 0.5       ],
               [ 0.5       ]])

        Multivariate Setting:

        >>> data_matrix = [[[[1, 3], [2, 6]], [[23, 54], [43, 76]], [[2, 45], [12, 65]]],
        ...                [[[21, 34], [8, 16]], [[67, 43], [32, 21]], [[10, 24], [3, 12]]],
        ...                [[[4, 6], [4, 10]], [[45, 48], [38, 56]], [[8, 36], [10, 28]]]]
        >>> sample_points = [[2, 4, 6], [3, 6]]
        >>> fd = FDataGrid(data_matrix, sample_points)
        >>> band_depth(fd)
        array([[ 0.66666667,  0.66666667],
               [ 0.66666667,  0.66666667],
               [ 1.        ,  1.        ]])

    """
    n = fdatagrid.nsamples
    nchoose2 = n * (n - 1) / 2

    ranks = _rank_samples(fdatagrid)
    axis = tuple(range(1, fdatagrid.ndim_domain + 1))
    nsamples_above = fdatagrid.nsamples - np.amax(ranks, axis=axis)
    nsamples_below = np.amin(ranks, axis=axis) - 1
    depth = (nsamples_below * nsamples_above + fdatagrid.nsamples - 1) / nchoose2

    if pointwise:
        _, depth_pointwise = modified_band_depth(fdatagrid, pointwise)
        return depth, depth_pointwise
    else:
        return depth


def modified_band_depth(fdatagrid, pointwise=False):
    """Implementation of Modified Band Depth for functional data.

    The band depth of each sample is obtained by computing the fraction of time its graph is contained
    in the bands determined by two sample curves. In the case the fdatagrid domain dimension is 2, instead
    of curves, surfaces determine the bands. In larger dimensions, the hyperplanes determine the bands.

    Args:
        fdatagrid (FDataGrid): Object over whose samples the modified band depth is going to be calculated.
        pointwise (boolean, optional): Indicates if the pointwise depth is also returned. Defaults to False.

    Returns:
        depth (numpy.darray): Array containing the modified band depth of the samples.

    Returns:
        depth_pointwise (numpy.darray, optional): Array containing the modified band depth of
        the samples at each point of discretisation. Only returned if pointwise equals to True.

    Examples:
        Univariate setting specifying pointwise:

        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2], [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5], [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> sample_points = [0, 2, 4, 6, 8, 10]
        >>> fd = FDataGrid(data_matrix, sample_points)
        >>> modified_band_depth(fd, pointwise = True)
        (array([[ 0.5       ],
               [ 0.83333333],
               [ 0.72222222],
               [ 0.66666667]]), array([[[ 0.5       ],
                [ 0.5       ],
                [ 0.5       ],
                [ 0.5       ],
                [ 0.5       ],
                [ 0.5       ]],
        <BLANKLINE>
               [[ 0.83333333],
                [ 0.83333333],
                [ 0.83333333],
                [ 0.83333333],
                [ 0.83333333],
                [ 0.83333333]],
        <BLANKLINE>
               [[ 0.5       ],
                [ 0.5       ],
                [ 0.83333333],
                [ 0.83333333],
                [ 0.83333333],
                [ 0.83333333]],
        <BLANKLINE>
               [[ 0.83333333],
                [ 0.83333333],
                [ 0.83333333],
                [ 0.5       ],
                [ 0.5       ],
                [ 0.5       ]]]))

        Multivariate Setting without specifying pointwise:

        >>> data_matrix = [[[[1, 3], [2, 6]], [[23, 54], [43, 76]], [[2, 45], [12, 65]]],
        ...                [[[21, 34], [8, 16]], [[67, 43], [32, 21]], [[10, 24], [3, 12]]],
        ...                [[[4, 6], [4, 10]], [[45, 48], [38, 56]], [[34, 78], [10, 28]]]]
        >>> sample_points = [[2, 4, 6], [3, 6]]
        >>> fd = FDataGrid(data_matrix, sample_points)
        >>> modified_band_depth(fd)
        array([[ 0.66666667,  0.72222222],
               [ 0.72222222,  0.66666667],
               [ 0.94444444,  0.94444444]])

    """
    n = fdatagrid.nsamples
    nchoose2 = n * (n - 1) / 2

    ranks = _rank_samples(fdatagrid)
    nsamples_above = fdatagrid.nsamples - ranks
    nsamples_below = ranks - 1
    match = nsamples_above * nsamples_below
    axis = tuple(range(1, fdatagrid.ndim_domain + 1))
    npoints_sample = reduce(lambda x, y: x * len(y), fdatagrid.sample_points, 1)
    proportion = match.sum(axis=axis) / npoints_sample
    depth = (proportion + fdatagrid.nsamples - 1) / nchoose2

    if pointwise:
        depth_pointwise = (match + fdatagrid.nsamples - 1) / nchoose2
        return depth, depth_pointwise
    else:
        return depth


def _cumulative_distribution(column):
    """Calculates the cumulative distribution function of the values passed to the function and evaluates it at each point.

    Args:
        column (numpy.darray): Array containing the values over which the distribution function is calculated.

    Returns:
        numpy.darray: Array containing the evaluation at each point of the distribution function.

    Examples:
        >>> _cumulative_distribution(np.array([1, 4, 5, 1, 2, 2, 4, 1, 1, 3]))
        array([ 0.4,  0.9,  1. ,  0.4,  0.6,  0.6,  0.9,  0.4,  0.4,  0.7])

    """
    if len(column.shape) != 1:
        raise ValueError("Only supported 1 dimensional arrays.")
    _, indexes, counts = np.unique(column, return_inverse=True, return_counts=True)
    count_cumulative = np.cumsum(counts) / len(column)
    return count_cumulative[indexes].reshape(column.shape)


def fraiman_muniz_depth(fdatagrid, pointwise=False):
    r"""Implementation of Fraiman and Muniz (FM) Depth for functional data.

    Each column is considered as the samples of an aleatory variable. The univariate depth of each of the samples of each column is
    calculated as follows:

    .. math::
        D(x) = 1 - \left\lvert \frac{1}{2}- F(x)\right\rvert

    Where :math:`F` stands for the marginal univariate distribution function of each column.

    The depth of a sample is the result of adding the previously computed depth for each of its points.

    Args:
        fdatagrid (FDataGrid): Object over whose samples the FM depth is going to be calculated.
        pointwise (boolean, optional): Indicates if the pointwise depth is also returned. Defaults to False.

    Returns:
        depth (numpy.darray): Array containing the FM depth of the samples.

    Returns:
        depth_pointwise (numpy.darray, optional): Array containing the FM depth of
        the samples at each point of discretisation. Only returned if pointwise equals to True.

    Examples:
        Univariate setting specifying pointwise:

        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2], [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5], [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> sample_points = [0, 2, 4, 6, 8, 10]
        >>> fd = FDataGrid(data_matrix, sample_points)
        >>> fraiman_muniz_depth(fd, pointwise = True)
        (array([[ 0.5       ],
               [ 0.75      ],
               [ 0.91666667],
               [ 0.875     ]]), array([[[ 0.5 ],
                [ 0.5 ],
                [ 0.5 ],
                [ 0.5 ],
                [ 0.5 ],
                [ 0.5 ]],
        <BLANKLINE>
               [[ 0.75],
                [ 0.75],
                [ 0.75],
                [ 0.75],
                [ 0.75],
                [ 0.75]],
        <BLANKLINE>
               [[ 0.75],
                [ 0.75],
                [ 1.  ],
                [ 1.  ],
                [ 1.  ],
                [ 1.  ]],
        <BLANKLINE>
               [[ 1.  ],
                [ 1.  ],
                [ 1.  ],
                [ 0.75],
                [ 0.75],
                [ 0.75]]]))

        Multivariate Setting without specifying pointwise:

        >>> data_matrix = [[[[1, 3], [2, 6]], [[23, 54], [43, 76]], [[2, 45], [12, 65]]],
        ...                [[[21, 34], [8, 16]], [[67, 43], [32, 21]], [[10, 24], [3, 12]]],
        ...                [[[4, 6], [4, 10]], [[45, 48], [38, 56]], [[34, 78], [10, 28]]]]
        >>> sample_points = [[2, 4, 6], [3, 6]]
        >>> fd = FDataGrid(data_matrix, sample_points)
        >>> fraiman_muniz_depth(fd)
        array([[ 0.72222222,  0.66666667],
               [ 0.66666667,  0.72222222],
               [ 0.77777778,  0.77777778]])

    """
    univariate_depth = np.zeros(fdatagrid.shape)

    ncols_dim_image = np.asarray([range(fdatagrid.shape[i]) for i in range(len(fdatagrid.shape) - 1, 0, -1)])
    tuples = list(itertools.product(*ncols_dim_image))
    for t in tuples:
        column = fdatagrid.data_matrix.T[t]
        univariate_depth.T[t] = 1 - np.abs(0.5 - _cumulative_distribution(column))

    axis = tuple(range(1, fdatagrid.ndim_domain + 1))
    npoints_sample = reduce(lambda x, y: x * len(y), fdatagrid.sample_points, 1)

    if pointwise:
        return np.sum(univariate_depth, axis=axis) / npoints_sample, univariate_depth
    else:
        return np.sum(univariate_depth, axis=axis) / npoints_sample



import unittest

import numpy as np

import matplotlib.pyplot as plt

from skfda import FDataGrid
from skfda.representation.interpolation import SplineInterpolator
from skfda.representation.basis import Fourier
from skfda.datasets import (make_multimodal_samples, make_multimodal_landmarks,
                            make_sinusoidal_process)
from skfda.preprocessing.registration import (
    normalize_warping, invert_warping, landmark_shift_deltas, landmark_shift,
    landmark_registration_warping, landmark_registration, mse_decomposition,
    shift_registration_deltas, shift_registration)


class TestWarping(unittest.TestCase):
    """Test warpings functions"""

    def setUp(self):
        """Initialization of samples"""

        self.time = np.linspace(-1, 1, 50)
        interpolator = SplineInterpolator(3, monotone=True)
        self.polynomial = FDataGrid([self.time**3, self.time**5],
                                    self.time, interpolator=interpolator)

    def test_invert_warping(self):

        inverse = invert_warping(self.polynomial)

        # Check if identity
        id = self.polynomial.compose(inverse)

        np.testing.assert_array_almost_equal([self.time, self.time],
                                             id.data_matrix[..., 0],
                                             decimal=3)

    def test_standard_normalize_warping(self):
        """Test normalization to (0, 1)"""

        normalized = normalize_warping(self.polynomial, (0, 1))

        # Test new domain range (0, 1)
        np.testing.assert_array_equal(normalized.domain_range, [(0, 1)])

        np.testing.assert_array_almost_equal(normalized.sample_points[0],
                                             np.linspace(0, 1, 50))

        np.testing.assert_array_almost_equal(normalized(0), [[0.], [0.]])

        np.testing.assert_array_almost_equal(normalized(1), [[1.], [1.]])

    def test_normalize_warpig(self):
        """Test normalization to (a, b)"""
        a = -4
        b = 3
        domain = (a, b)
        normalized = normalize_warping(self.polynomial, domain)

        # Test new domain range (0, 1)
        np.testing.assert_array_equal(normalized.domain_range, [domain])

        np.testing.assert_array_almost_equal(normalized.sample_points[0],
                                             np.linspace(*domain, 50))

        np.testing.assert_array_equal(normalized(a), [[a], [a]])

        np.testing.assert_array_equal(normalized(b), [[b], [b]])

    def test_landmark_shift_deltas(self):

        fd = make_multimodal_samples(n_samples=3, random_state=1)
        landmarks = make_multimodal_landmarks(n_samples=3, random_state=1)
        landmarks = landmarks.squeeze()

        shifts = landmark_shift_deltas(fd, landmarks).round(3)
        np.testing.assert_almost_equal(shifts, [0.25, -0.25, -0.231])

    def test_landmark_shift(self):

        fd = make_multimodal_samples(n_samples=3, random_state=1)
        landmarks = make_multimodal_landmarks(n_samples=3, random_state=1)
        landmarks = landmarks.squeeze()

        original_modes = fd(landmarks.reshape((3, 1, 1)),
                            aligned_evaluation=False)
        # Test default location
        fd_registered = landmark_shift(fd, landmarks)
        center = (landmarks.max() + landmarks.min())/2
        reg_modes = fd_registered(center)

        # Test callable location
        np.testing.assert_almost_equal(reg_modes, original_modes, decimal=2)

        fd_registered = landmark_shift(fd, landmarks, location=np.mean)
        center = np.mean(landmarks)
        reg_modes = fd_registered(center)

        np.testing.assert_almost_equal(reg_modes, original_modes, decimal=2)

        # Test integer location
        fd_registered = landmark_shift(fd, landmarks, location=0)
        center = np.mean(landmarks)
        reg_modes = fd_registered(0)

        np.testing.assert_almost_equal(reg_modes, original_modes, decimal=2)

        # Test array location
        fd_registered = landmark_shift(fd, landmarks, location=[0, 0.1, 0.2])
        reg_modes = fd_registered([[0], [.1], [.2]], aligned_evaluation=False)

        np.testing.assert_almost_equal(reg_modes, original_modes, decimal=2)

    def test_landmark_registration_warping(self):
        fd = make_multimodal_samples(n_samples=3, n_modes=2, random_state=9)
        landmarks = make_multimodal_landmarks(n_samples=3, n_modes=2,
                                              random_state=9)
        landmarks = landmarks.squeeze()

        # Default location
        warping = landmark_registration_warping(fd, landmarks)
        center = (landmarks.max(axis=0) + landmarks.min(axis=0)) / 2
        np.testing.assert_almost_equal(warping(center), landmarks, decimal=1)

        # Fixed location
        center = [.3, .6]
        warping = landmark_registration_warping(fd, landmarks, location=center)
        np.testing.assert_almost_equal(warping(center), landmarks, decimal=3)

    def test_landmark_registration(self):
        fd = make_multimodal_samples(n_samples=3, n_modes=2, random_state=9)
        landmarks = make_multimodal_landmarks(n_samples=3, n_modes=2,
                                              random_state=9)
        landmarks = landmarks.squeeze()

        original_values = fd(landmarks.reshape(3, 2), aligned_evaluation=False)

        # Default location
        fd_reg = landmark_registration(fd, landmarks)
        center = (landmarks.max(axis=0) + landmarks.min(axis=0)) / 2
        np.testing.assert_almost_equal(fd_reg(center), original_values,
                                       decimal=2)

        # Fixed location
        center = [.3, .6]
        fd_reg = landmark_registration(fd, landmarks, location=center)
        np.testing.assert_array_almost_equal(fd_reg(center), original_values,
                                             decimal=2)

    def test_mse_decomposition(self):
        fd = make_multimodal_samples(n_samples=3, random_state=1)
        landmarks = make_multimodal_landmarks(n_samples=3, random_state=1)
        landmarks = landmarks.squeeze()
        warping = landmark_registration_warping(fd, landmarks)
        fd_registered = fd.compose(warping)
        ret = mse_decomposition(fd, fd_registered, warping)

        np.testing.assert_almost_equal(ret.mse_amp, 0.0009866997121476962)
        np.testing.assert_almost_equal(ret.mse_pha, 0.11576861468435257)
        np.testing.assert_almost_equal(ret.rsq, 0.9915489952877273)
        np.testing.assert_almost_equal(ret.cr, 0.9999963424653829)

    def test_shift_registration_deltas(self):

        fd = make_sinusoidal_process(n_samples=2, error_std=0, random_state=1)

        deltas = shift_registration_deltas(fd).round(3)
        np.testing.assert_array_almost_equal(deltas, [-0.022,  0.03])

        fd = fd.to_basis(Fourier())
        deltas = shift_registration_deltas(fd).round(3)
        np.testing.assert_array_almost_equal(deltas, [-0.022,  0.03])

    def test_shift_registration(self):
        """Test wrapper of shift_registration_deltas"""

        fd = make_sinusoidal_process(n_samples=2, error_std=0, random_state=1)

        fd_reg = shift_registration(fd)
        deltas = shift_registration_deltas(fd)
        np.testing.assert_array_almost_equal(fd_reg.data_matrix,
                                             fd.shift(deltas).data_matrix)


if __name__ == '__main__':
    print()
    unittest.main()



"""K-Means Algorithms Module."""

import numpy as np
from ...representation.grid import FDataGrid
from ...misc.metrics import pairwise_distance, lp_distance
from abc import abstractmethod
from sklearn.base import BaseEstimator, ClusterMixin, TransformerMixin
from sklearn.exceptions import NotFittedError
import warnings
from sklearn.utils import check_random_state

__author__ = "Amanda Hernando Bernabé"
__email__ = "amanda.hernando@estudiante.uam.es"


class BaseKMeans(BaseEstimator, ClusterMixin, TransformerMixin):
    """Base class to implement K-Means clustering algorithms.

    Class from which both :class:`K-Means
    <skfda.ml.clustering.base_kmeans.KMeans>` and
    :class:`Fuzzy K-Means <skfda.ml.clustering.base_kmeans.FuzzyKMeans>`
    classes inherit.
    """

    def __init__(self, n_clusters, init, metric, n_init, max_iter, tol,
                 random_state):
        """Initialization of the BaseKMeans class.

        Args:
            n_clusters (int, optional): Number of groups into which the samples are
                classified. Defaults to 2.
            init (FDataGrid, optional): Contains the initial centers of the different
                clusters the algorithm starts with. Its data_marix must be of
                the shape (n_clusters, fdatagrid.ncol, fdatagrid.ndim_image).
                Defaults to None, and the centers are initialized randomly.
            metric (optional): metric that acceps two FDataGrid objects and returns
                a matrix with shape (fdatagrid1.nsamples, fdatagrid2.nsamples).
                Defaults to *pairwise_distance(lp_distance)*.
            n_init (int, optional): Number of time the k-means algorithm will be
                run with different centroid seeds. The final results will be the
                best output of n_init consecutive runs in terms of inertia.
            max_iter (int, optional): Maximum number of iterations of the
                clustering algorithm for a single run. Defaults to 100.
            tol (float, optional): tolerance used to compare the centroids
                calculated with the previous ones in every single run of the
                algorithm.
            random_state (int, RandomState instance or None, optional):
                Determines random number generation for centroid initialization. ç
                Use an int to make the randomness deterministic. Defaults to 0.
                See :term:`Glossary <random_state>`.
        """
        self.n_clusters = n_clusters
        self.init = init
        self.metric = metric
        self.n_init = n_init
        self.max_iter = max_iter
        self.tol = tol
        self.random_state = random_state

    def _generic_clustering_checks(self, fdatagrid):
        """Checks the arguments used in the
        :func:`fit method <skfda.ml.clustering.base_kmeans.fit>`.

        Args:
            fdatagrid (FDataGrid object): Object whose samples
                are classified into different groups.
        """

        if fdatagrid.ndim_domain > 1:
            raise NotImplementedError(
                "Only support 1 dimension on the domain.")

        if fdatagrid.nsamples < 2:
            raise ValueError(
                "The number of observations must be greater than 1.")

        if self.n_clusters < 2:
            raise ValueError(
                "The number of clusters must be greater than 1.")

        if self.n_init < 1:
            raise ValueError(
                "The number of iterations must be greater than 0.")

        if self.init is not None and self.n_init != 1:
            self.n_init = 1
            warnings.warn("Warning: The number of iterations is ignored "
                          "because the init parameter is set.")

        if self.init is not None and self.init.shape != (
                self.n_clusters, fdatagrid.ncol, fdatagrid.ndim_image):
            raise ValueError(
                "The init FDataGrid data_matrix should be of shape (n_clusters, "
                "n_features, ndim_image) and gives the initial centers.")

        if self.max_iter < 1:
            raise ValueError(
                "The number of maximum iterations must be greater than 0.")

        if self.tol < 0:
            raise ValueError("The tolerance must be positive.")

        return fdatagrid

    def _init_centroids(self, fdatagrid, random_state):
        """Compute the initial centroids

        Args:
            data_matrix (ndarray): matrix with the data only of the
                dimension of the image of the fdatagrid the algorithm is
                classifying.
            fdatagrid (FDataGrid object): Object whose samples are
                classified into different groups.
            random_state (RandomState object): random number generation for
                centroid initialization.

        Returns:
            centers (ndarray): initial centers
        """
        comparison = True
        while comparison:
            indices = random_state.permutation(fdatagrid.nsamples)[
                      :self.n_clusters]
            centers = fdatagrid.data_matrix[indices]
            unique_centers = np.unique(centers, axis=0)
            comparison = len(unique_centers) != self.n_clusters

        return centers

    @abstractmethod
    def fit(self, X, y=None, sample_weight=None):
        """ Computes clustering.

        Args:
            X (FDataGrid object): Object whose samples are clusered,
                classified into different groups.
             y (Ignored): present here for API consistency by convention.
            sample_weight (Ignored): present here for API consistency by
                convention.
        """
        pass

    def _check_is_fitted(self):
        """Perform is_fitted validation for estimator.

        Checks if the estimator is fitted by verifying the presence of
        of the calculated attributes "labels_" and "cluster_centers_", and
        raises a NotFittedError if that is not the case.
        """
        msg = ("This %(name)s instance is not fitted yet. Call 'fit' with "
               "appropriate arguments before using this method.")

        if not hasattr(self, "labels_") or \
                not hasattr(self, "cluster_centers_"):
            raise NotFittedError(msg % {'name': type(self).__name__})

    def _check_test_data(self, fdatagrid):
        """Checks that the FDataGrid object and the calculated centroids have
        compatible shapes.
        """
        if fdatagrid.shape[1:3] != self.cluster_centers_.shape[1:3]:
            raise ValueError("The fdatagrid shape is not the one expected for "
                             "the calculated cluster_centers_.")

    def predict(self, X, sample_weight=None):
        """Predict the closest cluster each sample in X belongs to.

        Args:
            X (FDataGrid object): Object whose samples are classified into
                different groups.
            y (Ignored): present here for API consistency by convention.
            sample_weight (Ignored): present here for API consistency by
                convention.

        Returns:
            labels_
        """
        self._check_is_fitted()
        self._check_test_data(X)
        return self.labels_

    def fit_predict(self, X, y=None, sample_weight=None):
        """Compute cluster centers and predict cluster index for each sample.

        Args:
            X (FDataGrid object): Object whose samples are classified into
                different groups.
            y (Ignored): present here for API consistency by convention.
            sample_weight (Ignored): present here for API consistency by
                convention.

        Returns:
            labels_
        """
        self.fit(X)
        return self.labels_

    def transform(self, X):
        """Transform X to a cluster-distance space.

        Args:
            X (FDataGrid object): Object whose samples are classified into
                different groups.
            y (Ignored): present here for API consistency by convention.
            sample_weight (Ignored): present here for API consistency by
                convention.

        Returns:
            distances_to_centers (numpy.ndarray: (nsamples, n_clusters)):
                distances of each sample to each cluster.
        """
        self._check_is_fitted()
        self._check_test_data(X)
        return self._distances_to_centers

    def fit_transform(self, X, y=None, sample_weight=None):
        """Compute clustering and transform X to cluster-distance space.

        Args:
            X (FDataGrid object): Object whose samples are classified into
                different groups.
            y (Ignored): present here for API consistency by convention.
            sample_weight (Ignored): present here for API consistency by
                convention.

        Returns:
            distances_to_centers (numpy.ndarray: (nsamples, n_clusters)):
                distances of each sample to each cluster.
        """
        self.fit(X)
        return self._distances_to_centers

    def score(self, X, y=None, sample_weight=None):
        """Opposite of the value of X on the K-means objective.

        Args:
            X (FDataGrid object): Object whose samples are classified into
                different groups.
            y (Ignored): present here for API consistency by convention.
            sample_weight (Ignored): present here for API consistency by
                convention.

        Returns:
            score (numpy.array: (fdatagrid.ndim_image)): negative *inertia_*
                attribute.

        """
        self._check_is_fitted()
        self._check_test_data(X)
        return -self.inertia_


class KMeans(BaseKMeans):
    r"""Representation and implementation of the K-Means algorithm
    for the FdataGrid object.

    Let :math:`\mathbf{X = \left\{ x_{1}, x_{2}, ..., x_{n}\right\}}` be a given dataset to be
    analyzed, and :math:`\mathbf{V = \left\{ v_{1}, v_{2}, ..., v_{c}\right\}}` be the set of
    centers of clusters in :math:`\mathbf{X}` dataset in :math:`m` dimensional space
    :math:`\left(\mathbb{R}^m \right)`. Where :math:`n` is the number of objects, :math:`m` is the
    number of features, and :math:`c` is the number of partitions or clusters.

    KM iteratively computes cluster centroids in order to minimize the sum with respect to the specified
    measure. KM algorithm aims at minimizing an objective function known as the squared error function given
    as follows:

    .. math::
        J_{KM}\left(\mathbf{X}; \mathbf{V}\right) = \sum_{i=1}^{c}\sum_{j=1}^{n}D_{ij}^2

    Where, :math:`D_{ij}^2` is the squared chosen distance measure which can be any p-norm:
    :math:`D_{ij} = \lVert x_{ij} - v_{i} \rVert = \left( \int_I \lvert x_{ij} - v_{i}\rvert^p dx \right)^{ \frac{1}{p}}`,
    being :math:`I` the domain where :math:`\mathbf{X}` is defined, :math:`1 \leqslant i \leqslant c`,
    :math:`1 \leqslant j\leqslant n_{i}`. Where :math:`n_{i}` represents the number of data points in i-th cluster.

    For :math:`c` clusters, KM is based on an iterative algorithm minimizing the sum of distances from each
    observation to its cluster centroid. The observations are moved between clusters until the sum cannot be decreased
    any more. KM algorithm involves the following steps:

    1. Centroids of :math:`c` clusters are chosen from :math:`\mathbf{X}` randomly or are passed to the
        function as a parameter.

    2. Distances between data points and cluster centroids are calculated.

    3. Each data point is assigned to the cluster whose centroid is closest to it.

    4. Cluster centroids are updated by using the following formula:
        :math:`\mathbf{v_{i}} ={\sum_{i=1}^{n_{i}}x_{ij}}/n_{i}` :math:`1 \leqslant i \leqslant c`.

    5. Distances from the updated cluster centroids are recalculated.

    6. If no data point is assigned to a new cluster the run of algorithm is stopped, otherwise the
        steps from 3 to 5 are repeated for probable movements of data points between the clusters.

    This algorithm is applied for each dimension on the image of the FDataGrid object.

    Args:
        n_clusters (int, optional): Number of groups into which the samples are 
            classified. Defaults to 2.
        init (FDataGrid, optional): Contains the initial centers of the different
            clusters the algorithm starts with. Its data_marix must be of
            the shape (n_clusters, fdatagrid.ncol, fdatagrid.ndim_image).
            Defaults to None, and the centers are initialized randomly.
        metric (optional): metric that acceps two FDataGrid objects and returns 
            a matrix with shape (fdatagrid1.nsamples, fdatagrid2.nsamples). 
            Defaults to *pairwise_distance(lp_distance)*.
        n_init (int, optional): Number of time the k-means algorithm will be 
            run with different centroid seeds. The final results will be the 
            best output of n_init consecutive runs in terms of inertia.
        max_iter (int, optional): Maximum number of iterations of the 
            clustering algorithm for a single run. Defaults to 100.
        tol (float, optional): tolerance used to compare the centroids 
            calculated with the previous ones in every single run of the 
            algorithm.
        random_state (int, RandomState instance or None, optional): 
            Determines random number generation for centroid initialization. ç
            Use an int to make the randomness deterministic. Defaults to 0.
            See :term:`Glossary <random_state>`.
            
    Attributes:
        labels_ (numpy.ndarray: (nsamples, ndim_image)): 2-dimensional matrix 
            in which each row contains the cluster that observation belongs to.
        cluster_centers_ (FDataGrid object): data_matrix of shape 
            (n_clusters, ncol, ndim_image) and contains the centroids for 
            each cluster.
        inertia_ (numpy.ndarray, (fdatagrid.ndim_image)): Sum of squared 
            distances of samples to their closest cluster center for each 
            dimension.
        n_iter_ (numpy.ndarray, (fdatagrid.ndim_image)): number of iterations the
            algorithm was run for each dimension.

    Example:

        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2], [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5], [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> sample_points = [0, 2, 4, 6, 8, 10]
        >>> fd = FDataGrid(data_matrix, sample_points)
        >>> kmeans = KMeans()
        >>> init= np.array([[0, 0, 0, 0, 0, 0], [2, 1, -1, 0.5, 0, -0.5]])
        >>> init_fd = FDataGrid(init, sample_points)
        >>> kmeans.fit(fd, init=init_fd)
        >>> kmeans
        KMeans(max_iter=100,
            metric=<function pairwise_distance.<locals>.pairwise at 0x7faf3aa061e0>, # doctest:+ELLIPSIS
            n_clusters=2, random_state=0, tol=0.0001)
    """.replace('+IGNORE_RESULT', '+ELLIPSIS\n<...>')

    def __init__(self, n_clusters=2, init=None,
                 metric=pairwise_distance(lp_distance),
                 n_init=1, max_iter=100, tol=1e-4, random_state=0):
        """Initialization of the KMeans class.

        Args:
            n_clusters (int, optional): Number of groups into which the samples are
                classified. Defaults to 2.
            init (FDataGrid, optional): Contains the initial centers of the different
                clusters the algorithm starts with. Its data_marix must be of
                the shape (n_clusters, fdatagrid.ncol, fdatagrid.ndim_image).
                Defaults to None, and the centers are initialized randomly.
            metric (optional): metric that acceps two FDataGrid objects and returns
                a matrix with shape (fdatagrid1.nsamples, fdatagrid2.nsamples).
                Defaults to *pairwise_distance(lp_distance)*.
            n_init (int, optional): Number of time the k-means algorithm will be
                run with different centroid seeds. The final results will be the
                best output of n_init consecutive runs in terms of inertia.
            max_iter (int, optional): Maximum number of iterations of the
                clustering algorithm for a single run. Defaults to 100.
            tol (float, optional): tolerance used to compare the centroids
                calculated with the previous ones in every single run of the
                algorithm.
            random_state (int, RandomState instance or None, optional):
                Determines random number generation for centroid initialization.
                Use an int to make the randomness deterministic. Defaults to 0.
        """
        super().__init__(n_clusters=n_clusters, init=init, metric=metric,
                         n_init=n_init, max_iter=max_iter, tol=tol,
                         random_state=random_state)

    def _kmeans_implementation(self, fdatagrid, random_state):
        """ Implementation of the K-Means algorithm for FDataGrid objects
        of any dimension.

        Args:
            fdatagrid (FDataGrid object): Object whose samples are clusered,
                classified into different groups.
            random_state (RandomState object): random number generation for
                centroid initialization.

        Returns:
            (tuple): tuple containing:

                clustering_values (numpy.ndarray: (nsamples,)): 1-dimensional
                array where each row contains the cluster that observation
                belongs to.

                centers (numpy.ndarray: (n_clusters, ncol, ndim_image)):
                Contains the centroids for each cluster.

                distances_to_centers (numpy.ndarray: (nsamples, n_clusters)):
                distances of each sample to each cluster.

                repetitions(int): number of iterations the algorithm was run.
        """
        repetitions = 0
        centers_old = np.zeros(
            (self.n_clusters, fdatagrid.ncol, fdatagrid.ndim_image))

        if self.init is None:
            centers = self._init_centroids(fdatagrid, random_state)
        else:
            centers = np.copy(self.init.data_matrix)

        while not np.allclose(centers, centers_old, rtol=self.tol,
                              atol=self.tol) and repetitions < self.max_iter:
            centers_old = np.copy(centers)
            centers_fd = FDataGrid(centers, fdatagrid.sample_points)
            distances_to_centers = self.metric(fdata1=fdatagrid,
                                               fdata2=centers_fd)
            clustering_values = np.argmin(distances_to_centers, axis=1)
            for i in range(self.n_clusters):
                indices, = np.where(clustering_values == i)
                if indices.size != 0:
                    centers[i] = np.average(
                        fdatagrid.data_matrix[indices, ...], axis=0)
            repetitions += 1

        return clustering_values, centers, distances_to_centers, repetitions

    def fit(self, X, y=None, sample_weight=None):
        """ Computes K-Means clustering calculating the attributes
        *labels_*, *cluster_centers_*, *inertia_* and *n_iter_*.

        Args:
            X (FDataGrid object): Object whose samples are clusered,
                classified into different groups.
             y (Ignored): present here for API consistency by convention.
            sample_weight (Ignored): present here for API consistency by
                convention.
        """
        random_state = check_random_state(self.random_state)
        fdatagrid = super()._generic_clustering_checks(fdatagrid=X)

        clustering_values = np.empty(
            (self.n_init, fdatagrid.nsamples)).astype(int)
        centers = np.empty((self.n_init, self.n_clusters,
                            fdatagrid.ncol, fdatagrid.ndim_image))
        distances_to_centers = np.empty(
            (self.n_init, fdatagrid.nsamples, self.n_clusters))
        distances_to_their_center = np.empty((self.n_init, fdatagrid.nsamples))
        n_iter = np.empty((self.n_init))

        for j in range(self.n_init):
            clustering_values[j, :], centers[j, :, :, :], \
            distances_to_centers[j, :, :], n_iter[j] = \
                self._kmeans_implementation(fdatagrid=fdatagrid,
                                            random_state=random_state)
            distances_to_their_center[j, :] = distances_to_centers[
                j, np.arange(fdatagrid.nsamples),
                clustering_values[j, :]]

        inertia = np.sum(distances_to_their_center ** 2, axis=1)
        index_best_iter = np.argmin(inertia)

        self.labels_ = clustering_values[index_best_iter]
        self.cluster_centers_ = FDataGrid(data_matrix=centers[index_best_iter],
                                          sample_points=fdatagrid.sample_points)
        self._distances_to_centers = distances_to_centers[index_best_iter]
        self.inertia_ = inertia[index_best_iter]
        self.n_iter_ = n_iter[index_best_iter]

        return self


class FuzzyKMeans(BaseKMeans):
    r""" Representation and implementation of the Fuzzy K-Means clustering
    algorithm for the FDataGrid object.

    Let :math:`\mathbf{X = \left\{ x_{1}, x_{2}, ..., x_{n}\right\}}` be a given dataset to be
    analyzed, and :math:`\mathbf{V = \left\{ v_{1}, v_{2}, ..., v_{c}\right\}}` be the set of
    centers of clusters in :math:`\mathbf{X}` dataset in :math:`m` dimensional space
    :math:`\left(\mathbb{R}^m \right)`. Where :math:`n` is the number of objects, :math:`m` is the
    number of features, and :math:`c` is the number of partitions or clusters.

    FCM minimizes the following objective function:

    .. math::
        J_{FCM}\left(\mathbf{X}; \mathbf{U, V}\right) = \sum_{i=1}^{c}\sum_{j=1}^{n}u_{ij}^{f}D_{ij}^2.

    This function differs from classical KM with the use of weighted squared errors instead of using squared
    errors only. In the objective function, :math:`\mathbf{U}` is a fuzzy partition matrix that is computed from
    dataset :math:`\mathbf{X}`: :math:`\mathbf{U} = [u_{ij}] \in M_{FCM}`.

    The fuzzy clustering of :math:`\mathbf{X}` is represented with :math:`\mathbf{U}` membership matrix. The element
    :math:`u_{ij}` is the membership value of j-th object to i-th cluster. In this case, the i-th row of :math:`\mathbf{U}`
    matrix is formed with membership values of :math:`n` objects to i-th cluster. :math:`\mathbf{V}` is a prototype vector
    of cluster prototypes (centroids): :math:`\mathbf{V = \left\{ v_{1}, v_{2}, ..., v_{c}\right\}}`,
    :math:`\mathbf{v_{i}}\in \mathbb{R}^m`.

    :math:`D_{ij}^2` is the squared chosen distance measure which can be any p-norm:
    :math:`D_{ij} =\lVert x_{ij} - v_{i} \rVert = \left( \int_I \lvert x_{ij} - v_{i}\rvert^p dx \right)^{ \frac{1}{p}}`,
    being :math:`I` the domain where :math:`\mathbf{X}` is defined, :math:`1 \leqslant i \leqslant c`,
    :math:`1 \leqslant j\leqslant n_{i}`. Where :math:`n_{i}` represents the number of data points in i-th cluster.

    FCM is an iterative process and stops when the number of iterations is reached to maximum, or when
    the centroids of the clusters do not change. The steps involved in FCM are:

        1. Centroids of :math:`c` clusters are chosen from :math:`\mathbf{X}` randomly or are passed to the
           function as a parameter.

        2. Membership values of data points to each cluster are calculated with:
           :math:`u_{ij} = \left[ \sum_{k=1}^c\left( D_{ij}/D_{kj}\right)^\frac{2}{f-1} \right]^{-1}`.

        3. Cluster centroids are updated by using the following formula:
           :math:`\mathbf{v_{i}} =\frac{\sum_{j=1}^{n}u_{ij}^f x_{j}}{\sum_{j=1}^{n} u_{ij}^f}`,
           :math:`1 \leqslant i \leqslant c`.

        4. If no cluster centroid changes the run of algorithm is stopped, otherwise return
           to step 2.

    This algorithm is applied for each dimension on the image of the FDataGrid object.

    Args:          
        n_clusters (int, optional): Number of groups into which the samples are 
            classified. Defaults to 2.
        init (FDataGrid, optional): Contains the initial centers of the different
            clusters the algorithm starts with. Its data_marix must be of
            the shape (n_clusters, fdatagrid.ncol, fdatagrid.ndim_image).
            Defaults to None, and the centers are initialized randomly.
        metric (optional): metric that acceps two FDataGrid objects and returns 
            a matrix with shape (fdatagrid1.nsamples, fdatagrid2.nsamples). 
            Defaults to *pairwise_distance(lp_distance)*.
        n_init (int, optional): Number of time the k-means algorithm will be 
            run with different centroid seeds. The final results will be the 
            best output of n_init consecutive runs in terms of inertia.
        max_iter (int, optional): Maximum number of iterations of the 
            clustering algorithm for a single run. Defaults to 100.
        tol (float, optional): tolerance used to compare the centroids 
            calculated with the previous ones in every single run of the 
            algorithm.
        random_state (int, RandomState instance or None, optional): 
            Determines random number generation for centroid initialization. ç
            Use an int to make the randomness deterministic. Defaults to 0.
            See :term:`Glossary <random_state>`.
        fuzzifier (int, optional): Scalar parameter used to specify the
            degree of fuzziness in the fuzzy algorithm. Defaults to 2.
        n_dec (int, optional): designates the number of decimals of the labels
            returned in the fuzzy algorithm. Defaults to 3.
            
    Attributes:
        labels_ (numpy.ndarray: (nsamples, ndim_image)): 2-dimensional matrix 
            in which each row contains the cluster that observation belongs to.
        cluster_centers_ (FDataGrid object): data_matrix of shape 
            (n_clusters, ncol, ndim_image) and contains the centroids for 
            each cluster.
        inertia_ (numpy.ndarray, (fdatagrid.ndim_image)): Sum of squared 
            distances of samples to their closest cluster center for each 
            dimension.
        n_iter_ (numpy.ndarray, (fdatagrid.ndim_image)): number of iterations the
            algorithm was run for each dimension.


    Example:

        >>> data_matrix = [[[1, 0.3], [2, 0.4], [3, 0.5], [4, 0.6]],
        ...                [[2, 0.5], [3, 0.6], [4, 0.7], [5, 0.7]],
        ...                [[3, 0.2], [4, 0.3], [5, 0.4], [6, 0.5]]]
        >>> sample_points = [2, 4, 6, 8]
        >>> fd = FDataGrid(data_matrix, sample_points)
        >>> fuzzy_kmeans = FuzzyKMeans()
        >>> init=np.array([[[3, 0], [5, 0], [2, 0], [4, 0]],
        ...                [[0, 0], [0, 1], [0, 0], [0, 1]]])
        >>> init_fd = FDataGrid(init, sample_points)
        >>> fuzzy_kmeans.fit(fd, init=init_fd)
        >>> fuzzy_kmeans
        FuzzyKMeans(fuzzifier=2, max_iter=100,
              metric=<function pairwise_distance.<locals>.pairwise at 0x7faf3aa06488>,  # doctest:+ELLIPSIS
              n_clusters=2, n_dec=3, random_state=0, tol=0.0001)
    """.replace('+IGNORE_RESULT', '+ELLIPSIS\n<...>')

    def __init__(self, n_clusters=2, init=None,
                 metric=pairwise_distance(lp_distance), n_init=1, max_iter=100,
                 tol=1e-4, random_state=0, fuzzifier=2, n_dec=3):
        """Initialization of the FuzzyKMeans class.

        Args:
            n_clusters (int, optional): Number of groups into which the samples are
                classified. Defaults to 2.
            init (FDataGrid, optional): Contains the initial centers of the different
                clusters the algorithm starts with. Its data_marix must be of
                the shape (n_clusters, fdatagrid.ncol, fdatagrid.ndim_image).
                Defaults to None, and the centers are initialized randomly.
            metric (optional): metric that acceps two FDataGrid objects and returns
                a matrix with shape (fdatagrid1.nsamples, fdatagrid2.nsamples).
                Defaults to *pairwise_distance(lp_distance)*.
            n_init (int, optional): Number of time the k-means algorithm will be
                run with different centroid seeds. The final results will be the
                best output of n_init consecutive runs in terms of inertia.
            max_iter (int, optional): Maximum number of iterations of the
                clustering algorithm for a single run. Defaults to 100.
            tol (float, optional): tolerance used to compare the centroids
                calculated with the previous ones in every single run of the
                algorithm.
            random_state (int, RandomState instance or None, optional):
                Determines random number generation for centroid initialization.
                Use an int to make the randomness deterministic. Defaults to 0.
            fuzzifier (int, optional): Scalar parameter used to specify the
                degree of fuzziness in the fuzzy algorithm. Defaults to 2.
            n_dec (int, optional): designates the number of decimals of the labels
                returned in the fuzzy algorithm. Defaults to 3.
        """
        super().__init__(n_clusters=n_clusters, init=init, metric=metric,
                         n_init=n_init,
                         max_iter=max_iter, tol=tol, random_state=random_state)

        self.fuzzifier = fuzzifier
        self.n_dec = n_dec

    def _fuzzy_kmeans_implementation(self, fdatagrid, random_state):
        """ Implementation of the Fuzzy K-Means algorithm for FDataGrid objects
        of any dimension.

        Args:
            fdatagrid (FDataGrid object): Object whose samples are clusered,
                classified into different groups.
            random_state (RandomState object): random number generation for
                centroid initialization.

        Returns:
            (tuple): tuple containing:

                membership values (numpy.ndarray: (nsamples, n_clusters)):
                2-dimensional matrix where each row contains the membership
                value that observation has to each cluster.

                centers (numpy.ndarray: (n_clusters, ncol, ndim_image)):
                Contains the centroids for each cluster.

                distances_to_centers (numpy.ndarray: (nsamples, n_clusters)):
                distances of each sample to each cluster.

                repetitions(int): number of iterations the algorithm was run.

        """
        repetitions = 0
        centers_old = np.zeros(
            (self.n_clusters, fdatagrid.ncol, fdatagrid.ndim_image))
        U = np.empty((fdatagrid.nsamples, self.n_clusters))
        distances_to_centers = np.empty((fdatagrid.nsamples, self.n_clusters))

        if self.init is None:
            centers = self._init_centroids(fdatagrid, random_state)
        else:
            centers = np.copy(self.init.data_matrix)

        while not np.allclose(centers, centers_old, rtol=self.tol,
                              atol=self.tol) and repetitions < self.max_iter:

            centers_old = np.copy(centers)
            centers_fd = FDataGrid(centers, fdatagrid.sample_points)
            distances_to_centers = self.metric(
                fdata1=fdatagrid,
                fdata2=centers_fd)
            distances_to_centers_raised = distances_to_centers ** (
                                2 / (self.fuzzifier - 1))

            for i in range(fdatagrid.nsamples):
                comparison = (fdatagrid.data_matrix[i] == centers).all(
                    axis=tuple(np.arange(fdatagrid.ndim)[1:]))
                if comparison.sum() >= 1:
                    U[i, np.where(comparison == True)] = 1
                    U[i, np.where(comparison == False)] = 0
                else:
                    for j in range(self.n_clusters):
                        U[i, j] = 1 / np.sum(
                            distances_to_centers_raised[i, j] /
                            distances_to_centers_raised[i])

            U = np.power(U, self.fuzzifier)
            for i in range(self.n_clusters):
                centers[i] = np.sum((U[:, i] * fdatagrid.data_matrix.T).T,
                                    axis=0) / np.sum(U[:, i])
            repetitions += 1

        return np.round(np.power(U, 1 / self.fuzzifier), self.n_dec), centers, \
               distances_to_centers, repetitions

    def fit(self, X, y=None, sample_weight=None):
        """ Computes Fuzzy K-Means clustering calculating the attributes
        *labels_*, *cluster_centers_*, *inertia_* and *n_iter_*.

        Args:
            X (FDataGrid object): Object whose samples are clusered,
                classified into different groups.
             y (Ignored): present here for API consistency by convention.
            sample_weight (Ignored): present here for API consistency by
                convention.
        """
        fdatagrid = super()._generic_clustering_checks(fdatagrid=X)
        random_state = check_random_state(self.random_state)

        if self.fuzzifier < 2:
            raise ValueError("The fuzzifier parameter must be greater than 1.")

        if self.n_dec < 1:
            raise ValueError(
                "The number of decimals should be greater than 0 in order to "
                "obtain a rational result.")

        membership_values = np.empty(
            (self.n_init, fdatagrid.nsamples, self.n_clusters))
        centers = np.empty(
            (self.n_init, self.n_clusters, fdatagrid.ncol,
             fdatagrid.ndim_image))
        distances_to_centers = np.empty(
            (self.n_init, fdatagrid.nsamples, self.n_clusters))
        distances_to_their_center = np.empty((self.n_init, fdatagrid.nsamples))
        n_iter = np.empty((self.n_init))

        for j in range(self.n_init):
            membership_values[j, :, :], centers[j, :, :, :], \
            distances_to_centers[j, :, :], n_iter[j] = \
                self._fuzzy_kmeans_implementation(fdatagrid=fdatagrid,
                                                  random_state=random_state)
            distances_to_their_center[j, :] = distances_to_centers[
                j, np.arange(fdatagrid.nsamples),
                np.argmax(membership_values[j, :, :], axis=-1)]

        inertia = np.sum(distances_to_their_center ** 2, axis=1)
        index_best_iter = np.argmin(inertia)

        self.labels_ = membership_values[index_best_iter]
        self.cluster_centers_ = FDataGrid(data_matrix=centers[index_best_iter],
                                          sample_points=fdatagrid.sample_points)
        self._distances_to_centers = distances_to_centers[index_best_iter]
        self.inertia_ = inertia[index_best_iter]
        self.n_iter_ = n_iter[index_best_iter]

        return self





import numpy as np
import scipy.integrate
import optimum_reparam
from . import invert_warping
from ._registration_utils import _normalize_scale

from ... import FDataGrid
from...representation.interpolation import SplineInterpolator


__author__ = "Pablo Marcos Manchón"
__email__ = "pablo.marcosm@estudiante.uam.es"

###############################################################################
# Based on the original implementation of J. Derek Tucker in                  #
# *fdasrsf_python* (https://github.com/jdtuck/fdasrsf_python)                 #
# and *ElasticFDA.jl* (https://github.com/jdtuck/ElasticFDA.jl).              #
###############################################################################


def to_srsf(fdatagrid, eval_points=None):
    """Calculate the square-root slope function (SRSF) transform.

    Let :math:`f_i : [a,b] \\rightarrow \\mathbb{R}` be an absolutely continuous
    function, the SRSF transform is defined as

    .. math::
        SRSF(f_i(t)) = sgn(f_i(t)) \\sqrt{|Df_i(t)|} = q_i(t)

    This representation it is used to compute the extended non-parametric
    Fisher-Rao distance between functions, wich under the SRSF representation
    becomes the usual :math:`\\mathbb{L}^2` distance between functions.
    See [SK16-4-6-1]_ .

    Args:
        fdatagrid (:class:`FDataGrid`): Functions to be transformed.
        eval_points: (array_like, optional): Set of points where the
            functions are evaluated, by default uses the sample points of the
            fdatagrid.

    Returns:
        :class:`FDataGrid`: SRSF functions.

    Raises:
        ValueError: If functions are multidimensional.

    References:
        ..  [SK16-4-6-1] Srivastava, Anuj & Klassen, Eric P. (2016). Functional
            and shape data analysis. In *Square-Root Slope Function
            Representation* (pp. 91-93). Springer.

    """

    if fdatagrid.ndim_domain > 1:
        raise ValueError("Only support functional objects with unidimensional "
                         "domain.")

    elif fdatagrid.ndim_image > 1:
        raise ValueError("Only support functional objects with unidimensional "
                         "image.")

    elif eval_points is None:
        eval_points = fdatagrid.sample_points[0]

    g = fdatagrid.derivative()

    # Evaluation with the corresponding interpolation
    g_data_matrix = g(eval_points, keepdims=False)

    # SRSF(f) = sign(f) * sqrt|Df|
    q_data_matrix = np.sign(g_data_matrix) * np.sqrt(np.abs(g_data_matrix))

    return fdatagrid.copy(data_matrix=q_data_matrix, sample_points=eval_points)


def from_srsf(fdatagrid, initial=None, *, eval_points=None):
    """Given a SRSF calculate the corresponding function in the original space.

    Let :math:`f_i : [a,b]\\rightarrow \\mathbb{R}` be an absolutely continuous
    function, the SRSF transform is defined as

    .. math::
        SRSF(f_i(t)) = sgn(f_i(t)) \\sqrt{|Df_i(t)|} = q_i(t)

    This transformation is a mapping up to constant. Given the srsf and the
    initial value the original function can be obtained as

    .. math::
        f_i(t) = f(a) + \\int_{a}^t q(t)|q(t)|dt

    This representation it is used to compute the extended non-parametric
    Fisher-Rao distance between functions, wich under the SRSF representation
    becomes the usual :math:`\\mathbb{L}^2` distance between functions.
    See [SK16-4-6-2]_ .

    Args:
        fdatagrid (:class:`FDataGrid`): SRSF to be transformed.
        initial (array_like): List of values of initial values of the original
            functions.
        eval_points: (array_like, optional): Set of points where the
            functions are evaluated, by default uses the sample points of the
            fdatagrid.

    Returns:
        :class:`FDataGrid`: Functions in the original space.

    Raises:
        ValueError: If functions are multidimensional.

    References:
        ..  [SK16-4-6-2] Srivastava, Anuj & Klassen, Eric P. (2016). Functional
            and shape data analysis. In *Square-Root Slope Function
            Representation* (pp. 91-93). Springer.

    """

    if fdatagrid.ndim_domain > 1:
        raise ValueError("Only support functional objects with "
                         "unidimensional domain.")

    elif fdatagrid.ndim_image > 1:
        raise ValueError("Only support functional objects with unidimensional "
                         "image.")

    elif eval_points is None:
        eval_points = fdatagrid.sample_points[0]

    q_data_matrix = fdatagrid(eval_points, keepdims=True)

    f_data_matrix = q_data_matrix * np.abs(q_data_matrix)

    f_data_matrix = scipy.integrate.cumtrapz(f_data_matrix,
                                             x=eval_points,
                                             axis=1,
                                             initial=0)

    if initial is not None:
        initial = np.atleast_1d(initial)
        initial = initial.reshape(fdatagrid.nsamples, 1, fdatagrid.ndim_image)
        initial = np.repeat(initial, len(eval_points), axis=1)
        f_data_matrix += initial

    return fdatagrid.copy(data_matrix=f_data_matrix, sample_points=eval_points)


def _elastic_alignment_array(template_data, q_data, eval_points, lam, grid_dim):
    """Wrapper between the cython interface and python.

    Selects the corresponding routine depending on the dimensions of the arrays.

    Args:
        template_data (numpy.ndarray): Array with the srsf of the template.
        q_data (numpy.ndarray): Array with the srsf of the curves to be aligned.
        eval_points (numpy.ndarray): Discretisation points of the functions.
        lam (float): Penalisation term.
        grid_dim (int): Dimension of the grid used in the alignment algorithm.

    Return:
        (numpy.ndarray): Array with the same shape than q_data with the srsf of
        the functions aligned to the template(s).
    """

    # Select cython function
    if template_data.ndim == 1 and q_data.ndim == 1:
        reparam = optimum_reparam.coptimum_reparam

    elif template_data.ndim == 1:
        reparam = optimum_reparam.coptimum_reparam_n

    else:
        reparam = optimum_reparam.coptimum_reparam_n2

    return reparam(np.ascontiguousarray(template_data.T),
                   np.ascontiguousarray(eval_points),
                   np.ascontiguousarray(q_data.T),
                   lam, grid_dim).T


def elastic_registration_warping(fdatagrid, template=None, *, lam=0.,
                                 eval_points=None, fdatagrid_srsf=None,
                                 template_srsf=None, grid_dim=7, **kwargs):
    """Calculate the warping to align a FDatagrid using the SRSF framework.

    Let :math:`f` be a function of the functional data object wich will be
    aligned to the template :math:`g`. Calculates the warping wich minimises
    the Fisher-Rao distance between :math:`g` and the registered function
    :math:`f^*(t)=f(\\gamma^*(t))=f \\circ \\gamma^*`.

    .. math::
        \\gamma^* = argmin_{\\gamma \\in \\Gamma} d_{\\lambda}(f \\circ
        \\gamma, g)

    Where :math:`d_{\\lambda}` denotes the extended amplitude distance with a
    penalty term, used to control the amount of warping.

    .. math::
        d_{\\lambda}^2(f \\circ \\gamma, g) = \\| SRSF(f \\circ \\gamma)
        \\sqrt{\\dot{\\gamma}} - SRSF(g)\\|_{\\mathbb{L}^2}^2 + \\lambda
        \\mathcal{R}(\\gamma)

    In the implementation it is used as penalty term

    .. math::
        \\mathcal{R}(\\gamma) = \\|\\sqrt{\\dot{\\gamma}}- 1 \\|_{\\mathbb{L}^2}^2

    Wich restrict the amount of elasticity employed in the alignment.

    The registered function :math:`f^*(t)` can be calculated using the
    composition :math:`f^*(t)=f(\\gamma^*(t))`.

    If the template is not specified it is used the Karcher mean of the set of
    functions under the Fisher-Rao metric to perform the alignment, wich is
    the local minimum of the sum of squares of elastic distances.
    See :func:`elastic_mean`.

    In [SK16-4-3]_ are described extensively the algorithms employed and the SRSF
    framework.

    Args:
        fdatagrid (:class:`FDataGrid`): Functional data object to be aligned.
        template (:class:`FDataGrid`, optional): Template to align the curves.
            Can contain 1 sample to align all the curves to it or the same
            number of samples than the fdatagrid. By default it is used the
            elastic mean.
        lam (float, optional): Controls the amount of elasticity. Defaults to 0.
        eval_points (array_like, optional): Set of points where the
            functions are evaluated, by default uses the sample points of the
            fdatagrid.
        fdatagrid_srsf (:class:`FDataGrid`, optional): SRSF of the fdatagrid,
            may be passed to avoid repeated calculation.
        template_srsf (:class:`FDataGrid`, optional): SRSF of the template,
            may be passed to avoid repeated calculation.
        grid_dim (int, optional): Dimension of the grid used in the alignment
            algorithm. Defaults 7.
        **kwargs: Named arguments to be passed to :func:`elastic_mean`.

    Returns:
        (:class:`FDataGrid`): Warping to align the given fdatagrid to the
        template.

    Raises:
        ValueError: If functions are multidimensional or the number of samples
            are different.

    References:
        ..  [SK16-4-3] Srivastava, Anuj & Klassen, Eric P. (2016). Functional
            and shape data analysis. In *Functional Data and Elastic
            Registration* (pp. 73-122). Springer.

    """

    # Check of params
    if fdatagrid.ndim_domain != 1 or fdatagrid.ndim_image != 1:

        raise ValueError("Not supported multidimensional functional objects.")

    if template is None:
        template = elastic_mean(fdatagrid, lam=lam, eval_points=eval_points,
                                **kwargs)

    elif ((template.nsamples != 1 and template.nsamples != fdatagrid.nsamples) or
          template.ndim_domain != 1 or template.ndim_image != 1):

        raise ValueError("The template should contain one sample to align all"
                         "the curves to the same function or the same number "
                         "of samples than the fdatagrid")

    # Construction of srsfs
    if fdatagrid_srsf is None:
        fdatagrid_srsf = to_srsf(fdatagrid, eval_points=eval_points)

    if template_srsf is None:
        template_srsf = to_srsf(template, eval_points=eval_points)

    if eval_points is None:
        eval_points = fdatagrid_srsf.sample_points[0]

    # Discretizacion in evaluation points
    q_data = fdatagrid_srsf(eval_points, keepdims=False).squeeze()
    template_data = template_srsf(eval_points, keepdims=False).squeeze()

    # Values of the warping
    gamma = _elastic_alignment_array(template_data, q_data,
                                     _normalize_scale(eval_points),
                                     lam, grid_dim)

    # Normalize warping to original interval
    gamma = _normalize_scale(gamma, a=eval_points[0], b=eval_points[-1])

    # Interpolator
    interpolator = SplineInterpolator(interpolation_order=3, monotone=True)

    return FDataGrid(gamma, eval_points, interpolator=interpolator)


def elastic_registration(fdatagrid, template=None, *, lam=0., eval_points=None,
                         fdatagrid_srsf=None, template_srsf=None, grid_dim=7,
                         **kwargs):
    """Align a FDatagrid using the SRSF framework.

    Let :math:`f` be a function of the functional data object wich will be
    aligned to the template :math:`g`. Calculates the warping wich minimises
    the Fisher-Rao distance between :math:`g` and the registered function
    :math:`f^*(t)=f(\\gamma^*(t))=f \\circ \\gamma^*`.

    .. math::
        \\gamma^* = argmin_{\\gamma \\in \\Gamma} d_{\\lambda}(f \\circ
        \\gamma, g)

    Where :math:`d_{\\lambda}` denotes the extended Fisher-Rao distance with a
    penalty term, used to control the amount of warping.

    .. math::
        d_{\\lambda}^2(f \\circ \\gamma, g) = \\| SRSF(f \\circ \\gamma)
        \\sqrt{\\dot{\\gamma}} - SRSF(g)\\|_{\\mathbb{L}^2}^2 + \\lambda
        \\mathcal{R}(\\gamma)

    In the implementation it is used as penalty term

    .. math::
        \\mathcal{R}(\\gamma) = \\|\\sqrt{\\dot{\\gamma}}- 1 \\|_{\\mathbb{L}^2}^2

    Wich restrict the amount of elasticity employed in the alignment.

    The registered function :math:`f^*(t)` can be calculated using the
    composition :math:`f^*(t)=f(\\gamma^*(t))`.

    If the template is not specified it is used the Karcher mean of the set of
    functions under the elastic metric to perform the alignment, wich is
    the local minimum of the sum of squares of elastic distances.
    See :func:`elastic_mean`.

    In [SK16-4-2]_ are described extensively the algorithms employed and the SRSF
    framework.

    Args:
        fdatagrid (:class:`FDataGrid`): Functional data object to be aligned.
        template (:class:`FDataGrid`, optional): Template to align the curves.
            Can contain 1 sample to align all the curves to it or the same
            number of samples than the fdatagrid. By default it is used the
            elastic mean.
        lam (float, optional): Controls the amount of elasticity. Defaults to 0.
        eval_points (array_like, optional): Set of points where the
            functions are evaluated, by default uses the sample points of the
            fdatagrid.
        fdatagrid_srsf (:class:`FDataGrid`, optional): SRSF of the fdatagrid,
            may be passed to avoid repeated calculation.
        template_srsf (:class:`FDataGrid`, optional): SRSF of the template,
            may be passed to avoid repeated calculation.
        grid_dim (int, optional): Dimension of the grid used in the alignment
            algorithm. Defaults 7.
        **kwargs: Named arguments to be passed to :func:`elastic_mean`.

    Returns:
        (:class:`FDataGrid`): FDatagrid with the samples aligned to the
            template.

    Raises:
        ValueError: If functions are multidimensional or the number of samples
            are different.

    References:
        ..  [SK16-4-2] Srivastava, Anuj & Klassen, Eric P. (2016). Functional
            and shape data analysis. In *Functional Data and Elastic
            Registration* (pp. 73-122). Springer.

    """

    # Calculates corresponding set of warpings
    warping = elastic_registration_warping(fdatagrid,
                                           template=template,
                                           lam=lam,
                                           eval_points=eval_points,
                                           fdatagrid_srsf=fdatagrid_srsf,
                                           template_srsf=template_srsf,
                                           grid_dim=grid_dim,
                                           **kwargs)

    return fdatagrid.compose(warping, eval_points=eval_points)


def warping_mean(warping, *, iter=20, tol=1e-5, step_size=1., eval_points=None,
                 return_shooting=False):
    """Compute the karcher mean of a set of warpings.

    Let :math:`\\gamma_i i=1...n` be a set of warping functions
    :math:`\\gamma_i:[a,b] \\rightarrow [a,b]` in :math:`\\Gamma`, i.e.,
    monotone increasing and with the restriction :math:`\\gamma_i(a)=a \\,
    \\gamma_i(b)=b`.

    The karcher mean :math:`\\bar \\gamma` is defined as the warping that
    minimises locally the sum of Fisher-Rao squared distances.
    [SK16-8-3-2]_.

    .. math::
        \\bar \\gamma = argmin_{\\gamma \\in \\Gamma} \\sum_{i=1}^{n}
         d_{FR}^2(\\gamma, \\gamma_i)

    The computation is performed using the structure of Hilbert Sphere obtained
    after a transformation of the warpings, see [S11-3-3]_.

    Args:
        warping (:class:`FDataGrid`): Set of warpings.
        iter (int): Maximun number of interations. Defaults to 20.
        tol (float): Convergence criterion, if the norm of the mean of the
            shooting vectors, :math:`| \\bar v |<tol`, the algorithm will stop.
            Defaults to 1e-5.
        step_size (float): Step size :math:`\\epsilon` used to update the mean.
            Default to 1.
        eval_points (array_like): Discretisation points of the warpings.
        shooting (boolean): If true it is returned a tuple with the mean and the
            shooting vectors, otherwise only the mean is returned.

    Return:
        (:class:`FDataGrid`) Fdatagrid with the mean of the warpings. If
        shooting is True the shooting vectors will be returned in a tuple with
        the mean.

    References:
        ..  [SK16-8-3-2] Srivastava, Anuj & Klassen, Eric P. (2016). Functional
            and shape data analysis. In *Template: Center of the Mean Orbit*
            (pp. 274-277). Springer.

        ..  [S11-3-3] Srivastava, Anuj et. al. Registration of Functional Data
            Using Fisher-Rao Metric (2011). In *Center of an Orbit* (pp. 9-10).
            arXiv:1103.3817v2.
    """

    if eval_points is None:
        eval_points = warping.sample_points[0]

    original_eval_points = eval_points

    if warping.sample_points[0][0] != 0 or warping.sample_points[0][-1] != 1:

        eval_points = _normalize_scale(eval_points)
        warping = FDataGrid(_normalize_scale(warping.data_matrix[..., 0]),
                            sample_points=_normalize_scale(warping.sample_points[0]))

    psi = to_srsf(warping, eval_points=eval_points).data_matrix[..., 0].T
    mu = to_srsf(warping.mean(), eval_points=eval_points).data_matrix[0]
    dot_aux = np.empty(psi.shape)

    n_points = mu.shape[0]

    sine = np.empty((warping.nsamples, 1))

    for _ in range(iter):
        # Dot product
        # <psi, mu> = S psi(t) mu(t) dt
        dot = scipy.integrate.simps(np.multiply(psi, mu, out=dot_aux),
                                    eval_points, axis=0)

        # Theorically is not possible (Cauchy–Schwarz inequallity), but due to
        # numerical approximation could be greater than 1
        dot[dot < -1] = -1
        dot[dot > 1] = 1
        theta = np.arccos(dot)[:, np.newaxis]

        # Be carefully with tangent vectors and division by 0
        idx = theta[:, 0] > tol
        sine[idx] = theta[idx] / np.sin(theta[idx])
        sine[~idx] = 0.

        # compute shooting vector
        cos_theta = np.repeat(np.cos(theta), n_points, axis=1)
        shooting = np.multiply(sine, (psi - np.multiply(cos_theta.T, mu)).T)

        # Mean of shooting vectors
        vmean = shooting.mean(axis=0, keepdims=True)
        v_norm = scipy.integrate.simps(np.square(vmean[0]))**(.5)

        # Convergence criterion
        if v_norm < tol:
            break

        # Update of mu
        mu *= np.cos(step_size*v_norm)
        vmean += np.sin(step_size * v_norm) / v_norm
        mu += vmean.T

    # Recover mean in original gamma space
    warping_mean = scipy.integrate.cumtrapz(np.square(mu, out=mu)[:, 0],
                                            x=eval_points, initial=0)

    # Affine traslation
    warping_mean = _normalize_scale(warping_mean,
                                    a=original_eval_points[0],
                                    b=original_eval_points[-1])

    monotone_interpolator = SplineInterpolator(interpolation_order=3,
                                               monotone=True)

    mean = FDataGrid([warping_mean], sample_points=original_eval_points,
                     interpolator=monotone_interpolator)

    # Shooting vectors are used in models based in the amplitude-phase
    # decomposition under this metric.
    if return_shooting:
        return mean, shooting

    return mean


def elastic_mean(fdatagrid, *, lam=0., center=True, iter=20, tol=1e-3,
                 initial=None, eval_points=None, fdatagrid_srsf=None,
                 grid_dim=7, **kwargs):
    """Compute the karcher mean under the elastic metric.

    Calculates the karcher mean of a set of functional samples in the amplitude
    space :math:`\\mathcal{A}=\\mathcal{F}/\\Gamma`.

    Let :math:`q_i` the corresponding SRSF of the observation :math:`f_i`.
    The space :math:`\\mathcal{A}` is defined using the equivalence classes
    :math:`[q_i]=\\{ q_i \\circ \\gamma \\| \\gamma \\in \\Gamma \\}`, where
    :math:`\\Gamma` denotes the space of warping functions. The karcher mean
    in this space is defined as

    .. math::
        [\\mu_q] = argmin_{[q] \\in \\mathcal{A}} \\sum_{i=1}^n
        d_{\\lambda}^2([q],[q_i])

    Once :math:`[\\mu_q]` is obtained it is selected the element of the
    equivalence class which makes the mean of the warpings employed be the
    identity.

    See [SK16-8-3-1]_ and [S11-3]_.

    Args:
        fdatagrid (:class:`FDataGrid`): Set of functions to compute the mean.
        lam (float): Penalisation term. Defaults to 0.
        center (boolean): If true it is computed the mean of the warpings and
            used to select a central mean. Defaults True.
        iter (int): Maximun number of iterations. Defaults to 20.
        tol (float): Convergence criterion, the algorithm will stop if
            :math:´\\|mu^{(\\nu)} - mu^{(\\nu - 1)} \\|_2 / \\| mu^{(\\nu-1)} \\|_2
            < tol´.
        initial (float): Value of the mean at the starting point. By default
            takes the average of the initial points of the samples.
        eval_points (array_like): Points of discretization of the fdatagrid.
        fdatagrid_srsf (:class:`FDataGrid`): SRSF if the fdatagrid, if it is
            passed it is not computed in the algorithm.
        grid_dim (int, optional): Dimension of the grid used in the alignment
            algorithm. Defaults 7.
        ** kwargs : Named options to be pased to :func:`warping_mean`.

    Return:
        (:class:`FDataGrid`): FDatagrid with the mean of the functions.

    Raises:
        ValueError: If the object is multidimensional or the shape of the srsf
            do not match with the fdatagrid.

    References:
        ..  [SK16-8-3-1] Srivastava, Anuj & Klassen, Eric P. (2016). Functional
            and shape data analysis. In *Karcher Mean of Amplitudes*
            (pp. 273-274). Springer.

        .. [S11-3] Srivastava, Anuj et. al. Registration of Functional Data
            Using Fisher-Rao Metric (2011). In *Karcher Mean and Function
            Alignment* (pp. 7-10). arXiv:1103.3817v2.

    """

    if fdatagrid.ndim_domain != 1 or fdatagrid.ndim_image != 1:
        raise ValueError("Not supported multidimensional functional objects.")

    if fdatagrid_srsf is not None and (fdatagrid_srsf.ndim_domain != 1 or
                                       fdatagrid_srsf.ndim_image != 1):
        raise ValueError("Not supported multidimensional functional objects.")

    elif fdatagrid_srsf is None:
        fdatagrid_srsf = to_srsf(fdatagrid, eval_points=eval_points)

    if eval_points is not None:
        eval_points = np.asarray(eval_points)
    else:
        eval_points = fdatagrid.sample_points[0]

    eval_points_normalized = _normalize_scale(eval_points)
    y_scale = eval_points[-1] - eval_points[0]

    interpolator = SplineInterpolator(interpolation_order=3, monotone=True)

    # Discretisation points
    fdatagrid_normalized = FDataGrid(fdatagrid(eval_points) / y_scale,
                                     sample_points=eval_points_normalized)

    srsf = fdatagrid_srsf(eval_points, keepdims=False)

    # Initialize with function closest to the L2 mean with the L2 distance
    centered = (srsf.T - srsf.mean(axis=0, keepdims=True).T).T

    distances = scipy.integrate.simps(np.square(centered, out=centered),
                                      eval_points_normalized, axis=1)

    # Initialization of iteration
    mu = srsf[np.argmin(distances)]
    mu_aux = np.empty(mu.shape)
    mu_1 = np.empty(mu.shape)

    # Main iteration
    for _ in range(iter):

        gammas = _elastic_alignment_array(
            mu, srsf, eval_points_normalized, lam, grid_dim)
        gammas = FDataGrid(gammas, sample_points=eval_points_normalized,
                           interpolator=interpolator)

        fdatagrid_normalized = fdatagrid_normalized.compose(gammas)
        srsf = to_srsf(fdatagrid_normalized).data_matrix[..., 0]

        # Next iteration
        mu_1 = srsf.mean(axis=0, out=mu_1)

        # Convergence criterion
        mu_norm = np.sqrt(scipy.integrate.simps(np.square(mu, out=mu_aux),
                                                eval_points_normalized))

        mu_diff = np.sqrt(scipy.integrate.simps(np.square(mu - mu_1, out=mu_aux),
                                                eval_points_normalized))

        if mu_diff / mu_norm < tol:
            break

        mu = mu_1

    if initial is None:
        initial = fdatagrid.data_matrix[:, 0].mean()

    # Karcher mean orbit in space L2/Gamma
    karcher_mean = from_srsf(fdatagrid.copy(data_matrix=[mu],
                                            sample_points=eval_points),
                             initial=initial)

    if center:
        # Gamma mean in Hilbert Sphere
        mean_normalized = warping_mean(gammas, return_shooting=False, **kwargs)

        gamma_mean = FDataGrid(_normalize_scale(mean_normalized.data_matrix[..., 0],
                                                a=eval_points[0],
                                                b=eval_points[-1]),
                               sample_points=eval_points)

        gamma_inverse = invert_warping(gamma_mean)

        karcher_mean = karcher_mean.compose(gamma_inverse)

    # Return center of the orbit
    return karcher_mean



from ._real_datasets import (fdata_constructor, fetch_cran,
                             fetch_ucr,
                             fetch_phoneme, fetch_growth,
                             fetch_tecator, fetch_medflies,
                             fetch_weather, fetch_aemet)
from ._samples_generators import (make_gaussian_process,
                                  make_sinusoidal_process,
                                  make_multimodal_samples,
                                  make_multimodal_landmarks,
                                  make_random_warping)



"""Landmark Registration of functional data module.

This module contains methods to perform the landmark registration.
"""

import numpy

from ... import FDataGrid
from ...representation.interpolation import SplineInterpolator

__author__ = "Pablo Marcos Manchón"
__email__ = "pablo.marcosm@estudiante.uam.es"


def landmark_shift_deltas(fd, landmarks, location=None):
    r"""Returns the corresponding shifts to align the landmarks of the curves.

        Let :math:`t^*` the time where the landmarks of the curves will be
        aligned, and :math:`t_i` the location of the landmarks for each curve.
        The function will calculate the corresponding :math:`\delta_i` shuch
        that :math:`t_i = t^* + \delta_i`.

        This procedure will work independent of the dimension of the domain
        and the image.

    Args:
        fd (:class:`FData`): Functional data object.
        landmarks (array_like): List with the landmarks of the samples.
        location (numeric or callable, optional): Defines where
            the landmarks will be alligned. If a numer or list is passed the
            landmarks will be alligned to it. In case of a callable is
            passed the location will be the result of the the call, the
            function should be accept as an unique parameter a numpy array
            with the list of landmarks.
            By default it will be used as location :math:`\frac{1}{2}(max(
            \text{landmarks})+ min(\text{landmarks}))` wich minimizes the
            max shift.

    Returns:
        :class:`numpy.ndarray`: Array containing the corresponding shifts.

    Raises:
        ValueError: If the list of landmarks does not match with the number of
            samples.

    Examples:

        >>> from skfda.datasets import make_multimodal_landmarks
        >>> from skfda.datasets import make_multimodal_samples
        >>> from skfda.preprocessing.registration import landmark_shift_deltas

        We will create a data with landmarks as example

        >>> fd = make_multimodal_samples(n_samples=3, random_state=1)
        >>> landmarks = make_multimodal_landmarks(n_samples=3, random_state=1)
        >>> landmarks = landmarks.squeeze()

        The function will return the corresponding shifts

        >>> shifts = landmark_shift_deltas(fd, landmarks)
        >>> shifts.round(3)
        array([ 0.25 , -0.25 , -0.231])

        The registered samples can be obtained with a shift

        >>> fd.shift(shifts)
        FDataGrid(...)

    """

    if len(landmarks) != fd.nsamples:
        raise ValueError(f"landmark list ({len(landmarks)}) must have the same "
                         f"length than the number of samples ({fd.nsamples})")

    landmarks = numpy.atleast_1d(landmarks)

    # Parses location
    if location is None:
        p = (numpy.max(landmarks, axis=0) + numpy.min(landmarks, axis=0)) / 2.
    elif callable(location):
        p = location(landmarks)
    else:
        try:
            p = numpy.atleast_1d(location)
        except:
            raise ValueError("Invalid location, must be None, a callable or a "
                             "number in the domain")

    shifts = landmarks - p

    return shifts


def landmark_shift(fd, landmarks, location=None, *, restrict_domain=False,
                   extrapolation=None, eval_points=None, **kwargs):
    r"""Perform a shift of the curves to align the landmarks.

        Let :math:`t^*` the time where the landmarks of the curves will be
        aligned, :math:`t_i` the location of the landmarks for each curve
        and :math:`\delta_i= t_i - t^*`.

        The registered samples will have their feature aligned.

        .. math::
            x_i^*(t^*)=x_i(t^* + \delta_i)=x_i(t_i)

    Args:
        fd (:class:`FData`): Functional data object.
        landmarks (array_like): List with the landmarks of the samples.
        location (numeric or callable, optional): Defines where
            the landmarks will be alligned. If a numeric value is passed the
            landmarks will be alligned to it. In case of a callable is
            passed the location will be the result of the the call, the
            function should be accept as an unique parameter a numpy array
            with the list of landmarks.
            By default it will be used as location :math:`\frac{1}{2}(max(
            \text{landmarks})+ min(\text{landmarks}))` wich minimizes the
            max shift.
        restrict_domain (bool, optional): If True restricts the domain to
            avoid evaluate points outside the domain using extrapolation.
            Defaults uses extrapolation.
        extrapolation (str or Extrapolation, optional): Controls the
            extrapolation mode for elements outside the domain range.
            By default uses the method defined in fd. See extrapolation to
            more information.
        eval_points (array_like, optional): Set of points where
            the functions are evaluated in :func:`shift`.
        **kwargs: Keyword arguments to be passed to :func:`shift`.

    Returns:
        :class:`FData`: Functional data object with the registered samples.

    Examples:

        >>> from skfda.datasets import make_multimodal_landmarks
        >>> from skfda.datasets import make_multimodal_samples
        >>> from skfda.preprocessing.registration import landmark_shift

        We will create a data with landmarks as example

        >>> fd = make_multimodal_samples(n_samples=3, random_state=1)
        >>> landmarks = make_multimodal_landmarks(n_samples=3, random_state=1)
        >>> landmarks = landmarks.squeeze()

        The function will return the sample registered

        >>> landmark_shift(fd, landmarks)
        FDataGrid(...)

    """

    shifts = landmark_shift_deltas(fd, landmarks, location=location)

    return fd.shift(shifts, restrict_domain=restrict_domain,
                    extrapolation=extrapolation,
                    eval_points=eval_points, **kwargs)


def landmark_registration_warping(fd, landmarks, *, location=None,
                                  eval_points=None):
    """Calculate the transformation used in landmark registration.

        Let :math:`t_{ij}` the time where the sample :math:`i` has the feature
        :math:`j` and :math:`t^*_j` the new time for the feature. The warping
        function will transform the new time in the old time, i.e.,
        :math:`h_i(t^*_j)=t_{ij}`.
        The registered samples can be obtained as :math:`x^*_i(t)=x_i(h_i(t))`.

        See [RS05-7-3-1]_ for a detailed explanation.

    Args:
        fd (:class:`FData`): Functional data object.
        landmarks (array_like): List containing landmarks for each samples.
        location (array_like, optional): Defines where
            the landmarks will be alligned. By default it will be used as
            location the mean of the landmarks.
        eval_points (array_like, optional): Set of points where
            the functions are evaluated to obtain a discrete
            representation of the object.
    Returns:
        :class:`FDataGrid`: FDataGrid with the warpings function needed to
        register the functional data object.

    Raises:
        ValueError: If the object to be registered has domain dimension greater
            than 1 or the list of landmarks or locations does not match with the
            number of samples.

    References:

    ..  [RS05-7-3-1] Ramsay, J., Silverman, B. W. (2005). Feature or landmark
        registration. In *Functional Data Analysis* (pp. 132-136). Springer.

    Examples:

        >>> from skfda.datasets import make_multimodal_landmarks
        >>> from skfda.datasets import make_multimodal_samples
        >>> from skfda.preprocessing.registration import landmark_registration_warping

        We will create a data with landmarks as example

        >>> fd = make_multimodal_samples(n_samples=3, n_modes=2, random_state=9)
        >>> landmarks = make_multimodal_landmarks(n_samples=3, n_modes=2,
        ...                                       random_state=9)
        >>> landmarks = landmarks.squeeze()

        The function will return the corresponding warping function

        >>> warping = landmark_registration_warping(fd, landmarks)
        >>> warping
        FDataGrid(...)

        The registered function can be obtained using function composition

        >>> fd.compose(warping)
        FDataGrid(...)
    """

    if fd.ndim_domain > 1:
        raise NotImplementedError("Method only implemented for objects with"
                                  "domain dimension up to 1.")

    if len(landmarks) != fd.nsamples:
        raise ValueError("The number of list of landmarks should be equal to "
                         "the number of samples")

    landmarks = numpy.asarray(landmarks).reshape((fd.nsamples, -1))

    n_landmarks = landmarks.shape[-1]

    data_matrix = numpy.empty((fd.nsamples, n_landmarks + 2))

    data_matrix[:, 0] = fd.domain_range[0][0]
    data_matrix[:, -1] = fd.domain_range[0][1]

    data_matrix[:, 1:-1] = landmarks

    if location is None:
        sample_points = numpy.mean(data_matrix, axis=0)

    elif n_landmarks != len(location):

        raise ValueError(f"Number of landmark locations should be equal than "
                         f"the number of landmarks ({len(location)}) != "
                         f"({n_landmarks})")
    else:
        sample_points = numpy.empty(n_landmarks + 2)
        sample_points[0] = fd.domain_range[0][0]
        sample_points[-1] = fd.domain_range[0][1]
        sample_points[1:-1] = location

    interpolator = SplineInterpolator(interpolation_order=3, monotone=True)

    warping = FDataGrid(data_matrix=data_matrix,
                        sample_points=sample_points,
                        interpolator=interpolator,
                        extrapolation='bounds')

    try:
        warping_points = fd.sample_points
    except AttributeError:
        warping_points = [numpy.linspace(*domain, 201)
                          for domain in fd.domain_range]

    return warping.to_grid(warping_points)


def landmark_registration(fd, landmarks, *, location=None, eval_points=None):
    """Perform landmark registration of the curves.

        Let :math:`t_{ij}` the time where the sample :math:`i` has the feature
        :math:`j` and :math:`t^*_j` the new time for the feature. The registered
        samples will have their features aligned, i.e.,
        :math:`x^*_i(t^*_j)=x_i(t_{ij})`.

        See [RS05-7-3]_ for a detailed explanation.

    Args:
        fd (:class:`FData`): Functional data object.
        landmarks (array_like): List containing landmarks for each samples.
        location (array_like, optional): Defines where
            the landmarks will be alligned. By default it will be used as
            location the mean of the landmarks.
        eval_points (array_like, optional): Set of points where
            the functions are evaluated to obtain a discrete
            representation of the object. In case of objects with
            multidimensional domain a list axis with points of evaluation
            for each dimension.

    Returns:
        :class:`FData`: FData with the functional data object registered.

    References:

    ..  [RS05-7-3] Ramsay, J., Silverman, B. W. (2005). Feature or landmark
        registration. In *Functional Data Analysis* (pp. 132-136). Springer.

    Examples:

        >>> from skfda.datasets import make_multimodal_landmarks
        >>> from skfda.datasets import make_multimodal_samples
        >>> from skfda.preprocessing.registration import landmark_registration
        >>> from skfda.representation.basis import BSpline

        We will create a data with landmarks as example

        >>> fd = make_multimodal_samples(n_samples=3, n_modes=2, random_state=9)
        >>> landmarks = make_multimodal_landmarks(n_samples=3, n_modes=2,
        ...                                       random_state=9)
        >>> landmarks = landmarks.squeeze()

        The function will return the registered curves

        >>> landmark_registration(fd, landmarks)
        FDataGrid(...)

        This method will work for FDataBasis as for FDataGrids

        >>> fd = fd.to_basis(BSpline(nbasis=12, domain_range=(-1,1)))
        >>> landmark_registration(fd, landmarks)
        FDataBasis(...)

    """

    warping = landmark_registration_warping(fd, landmarks, location=location,
                                            eval_points=eval_points)

    return fd.compose(warping)



"""Clustering Plots Module."""

from ...ml.clustering.base_kmeans import FuzzyKMeans
import numpy as np
import matplotlib.pyplot as plt
from mpldatacursor import datacursor
import matplotlib.patches as mpatches
from matplotlib.ticker import MaxNLocator
from sklearn.exceptions import NotFittedError
import warnings

__author__ = "Amanda Hernando Bernabé"
__email__ = "amanda.hernando@estudiante.uam.es"


def _check_if_estimator(estimator):
    """Checks the argument *estimator* is actually an estimator that
    implements the *fit* method.

    Args:
        estimator (BaseEstimator object): estimator used to calculate the
            clusters.
    """
    msg = ("This %(name)s instance has no attribute \"fit\".")
    if not hasattr(estimator, "fit"):
        raise AttributeError(msg % {'name': type(estimator).__name__})


def _plot_clustering_checks(estimator, fdatagrid, sample_colors, sample_labels,
                            cluster_colors, cluster_labels,
                            center_colors, center_labels):
    """Checks the arguments *sample_colors*, *sample_labels*, *cluster_colors*,
    *cluster_labels*, *center_colors*, *center_labels*, passed to the plot
    functions, have the correct dimensions.

    Args:
        estimator (BaseEstimator object): estimator used to calculate the
            clusters.
        fdatagrid (FDataGrd object): contains the samples which are grouped
            into different clusters.
        sample_colors (list of colors): contains in order the colors of each
            sample of the fdatagrid.
        sample_labels (list of str): contains in order the labels of each sample
            of the fdatagrid.
        cluster_colors (list of colors): contains in order the colors of each
            cluster the samples of the fdatagrid are classified into.
        cluster_labels (list of str): contains in order the names of each
            cluster the samples of the fdatagrid are classified into.
        center_colors (list of colors): contains in order the colors of each
            centroid of the clusters the samples of the fdatagrid are classified into.
        center_labels list of colors): contains in order the labels of each
            centroid of the clusters the samples of the fdatagrid are classified into.

    """

    if sample_colors is not None and len(
            sample_colors) != fdatagrid.nsamples:
        raise ValueError(
            "sample_colors must contain a color for each sample.")

    if sample_labels is not None and len(
            sample_labels) != fdatagrid.nsamples:
        raise ValueError(
            "sample_labels must contain a label for each sample.")

    if cluster_colors is not None and len(
            cluster_colors) != estimator.n_clusters:
        raise ValueError(
            "cluster_colors must contain a color for each cluster.")

    if cluster_labels is not None and len(
            cluster_labels) != estimator.n_clusters:
        raise ValueError(
            "cluster_labels must contain a label for each cluster.")

    if center_colors is not None and len(
            center_colors) != estimator.n_clusters:
        raise ValueError(
            "center_colors must contain a color for each center.")

    if center_labels is not None and len(
            center_labels) != estimator.n_clusters:
        raise ValueError(
            "centers_labels must contain a label for each center.")


def _plot_clusters(estimator, fdatagrid, fig, ax, nrows, ncols, labels,
                   sample_labels, cluster_colors, cluster_labels,
                   center_colors, center_labels, colormap):
    """Implementation of the plot of the FDataGrid samples by clusters.

    Args:
        estimator (BaseEstimator object): estimator used to calculate the
            clusters.
        fdatagrid (FDataGrd object): contains the samples which are grouped
            into different clusters.
        fig (figure object): figure over which the graphs are plotted in
            case ax is not specified. If None and ax is also None, the figure
            is initialized.
        ax (list of axis objects): axis over where the graphs are plotted.
            If None, see param fig.
        nrows(int): designates the number of rows of the figure to plot the
            different dimensions of the image. Only specified if fig and
            ax are None.
        ncols(int): designates the number of columns of the figure to plot
            the different dimensions of the image. Only specified if fig
            and ax are None.
        labels (numpy.ndarray, int: (nsamples, ndim_image)): 2-dimensional
            matrix where each row contains the number of cluster cluster
            that observation belongs to.
        sample_labels (list of str): contains in order the labels of each sample
            of the fdatagrid.
        cluster_colors (list of colors): contains in order the colors of each
            cluster the samples of the fdatagrid are classified into.
        cluster_labels (list of str): contains in order the names of each
            cluster the samples of the fdatagrid are classified into.
        center_colors (list of colors): contains in order the colors of each
            centroid of the clusters the samples of the fdatagrid are classified into.
        center_labels list of colors): contains in order the labels of each
            centroid of the clusters the samples of the fdatagrid are classified into.
        colormap(colormap): colormap from which the colors of the plot are taken.

    Returns:
        (tuple): tuple containing:

            fig (figure object): figure object in which the graphs are plotted in case ax is None.

            ax (axes object): axes in which the graphs are plotted.
    """
    fig, ax = fdatagrid.generic_plotting_checks(fig, ax, nrows, ncols)

    _plot_clustering_checks(estimator, fdatagrid, None, sample_labels,
                            cluster_colors, cluster_labels, center_colors,
                            center_labels)

    if sample_labels is None:
        sample_labels = ['$SAMPLE: {}$'.format(i) for i in
                         range(fdatagrid.nsamples)]

    if cluster_colors is None:
        cluster_colors = colormap(
            np.arange(estimator.n_clusters) / (estimator.n_clusters - 1))

    if cluster_labels is None:
        cluster_labels = ['$CLUSTER: {}$'.format(i) for i in
                          range(estimator.n_clusters)]

    if center_colors is None:
        center_colors = ["black"] * estimator.n_clusters

    if center_labels is None:
        center_labels = ['$CENTER: {}$'.format(i) for i in
                         range(estimator.n_clusters)]

    colors_by_cluster = cluster_colors[labels]

    patches = []
    for i in range(estimator.n_clusters):
        patches.append(
            mpatches.Patch(color=cluster_colors[i],
                           label=cluster_labels[i]))

    for j in range(fdatagrid.ndim_image):
        for i in range(fdatagrid.nsamples):
            ax[j].plot(fdatagrid.sample_points[0],
                       fdatagrid.data_matrix[i, :, j],
                       c=colors_by_cluster[i],
                       label=sample_labels[i])
        for i in range(estimator.n_clusters):
            ax[j].plot(fdatagrid.sample_points[0],
                       estimator.cluster_centers_.data_matrix[i, :, j],
                       c=center_colors[i], label=center_labels[i])
        ax[j].legend(handles=patches)
        datacursor(formatter='{label}'.format)

    fdatagrid.set_labels(fig, ax)

    return fig, ax


def plot_clusters(estimator, X, fig=None, ax=None, nrows=None, ncols=None,
                  sample_labels=None, cluster_colors=None,
                  cluster_labels=None, center_colors=None, center_labels=None,
                  colormap=plt.cm.get_cmap('rainbow')):
    """Plot of the FDataGrid samples by clusters.

    The clusters are calculated with the estimator passed as a parameter. If
    the estimator is not fitted, the fit method is called.
    Once each sample is assigned a label the plotting can be done.
    Each group is assigned a color described in a leglend.

    Args:
        estimator (BaseEstimator object): estimator used to calculate the
            clusters.
        X (FDataGrd object): contains the samples which are grouped
            into different clusters.
        fig (figure object): figure over which the graphs are plotted in
            case ax is not specified. If None and ax is also None, the figure
            is initialized.
        ax (list of axis objects): axis over where the graphs are plotted.
            If None, see param fig.
        nrows(int): designates the number of rows of the figure to plot the
            different dimensions of the image. Only specified if fig and
            ax are None.
        ncols(int): designates the number of columns of the figure to plot
            the different dimensions of the image. Only specified if fig
            and ax are None.
        sample_labels (list of str): contains in order the labels of each sample
            of the fdatagrid.
        cluster_colors (list of colors): contains in order the colors of each
            cluster the samples of the fdatagrid are classified into.
        cluster_labels (list of str): contains in order the names of each
            cluster the samples of the fdatagrid are classified into.
        center_colors (list of colors): contains in order the colors of each
            centroid of the clusters the samples of the fdatagrid are classified into.
        center_labels (list of colors): contains in order the labels of each
            centroid of the clusters the samples of the fdatagrid are classified into.
        colormap(colormap): colormap from which the colors of the plot are
            taken. Defaults to `rainbow`.

    Returns:
        (tuple): tuple containing:

            fig (figure object): figure object in which the graphs are plotted
            in case ax is None.

            ax (axes object): axes in which the graphs are plotted.
    """
    _check_if_estimator(estimator)
    try:
        estimator._check_is_fitted()
        estimator._check_test_data(X)
    except NotFittedError:
        estimator.fit(X)

    if isinstance(estimator, FuzzyKMeans):
        labels = np.argmax(estimator.labels_, axis=1)
    else:
        labels = estimator.labels_

    return _plot_clusters(estimator=estimator, fdatagrid=X,
                          fig=fig, ax=ax, nrows=nrows, ncols=ncols,
                          labels=labels, sample_labels=sample_labels,
                          cluster_colors=cluster_colors,
                          cluster_labels=cluster_labels,
                          center_colors=center_colors,
                          center_labels=center_labels, colormap=colormap)


def _set_labels(xlabel, ylabel, title, xlabel_str):
    """Sets the arguments *xlabel*, *ylabel*, *title* passed to the plot
    functions :func:`plot_cluster_lines
    <skfda.exploratory.visualization.clustering_plots.plot_cluster_lines>` and
    :func:`plot_cluster_bars <skfda.exploratory.visualization.clustering_plots.plot_cluster_bars>`,
    in case they are not set yet.

    Args:
        xlabel (lstr): Label for the x-axes.
        ylabel (str): Label for the y-axes.
        title (str): Title for the figure where the clustering results are ploted.
        xlabel_str (str): In case xlabel is None, string to use for the labels
            in the x-axes.

    Returns:
        xlabel (str): Labels for the x-axes.
        ylabel (str): Labels for the y-axes.
        title (str): Title for the figure where the clustering results are plotted.
    """
    if xlabel is None:
        xlabel = xlabel_str

    if ylabel is None:
        ylabel = "Membership grade"

    if title is None:
        title = "Membership grades of the samples to each cluster"

    return xlabel, ylabel, title

def _fig_and_ax_checks(fig, ax):
    """Checks the arguments *fig* and *ax* passed to the plot
    functions :func:`plot_cluster_lines
    <skfda.exploratory.visualization.clustering_plots.plot_cluster_lines>` and
    :func:`plot_cluster_bars <skfda.exploratory.visualization.clustering_plots.plot_cluster_bars>`.
    In case they are not set yet, they are initialised.

    Args:
        fig (figure object): figure over which the graph is
            plotted in case ax is not specified. If None and ax is also None,
            the figure is initialized.
        ax (axis object): axis over where the graph is  plotted.
            If None, see param fig.
    """
    if fig is not None and ax is not None:
        raise ValueError("fig and axes parameters cannot be passed as "
                         "arguments at the same time.")

    if fig is not None and len(fig.get_axes()) > 1:
        warnings.warn("Warning: The first axis of the figure is where "
                      "the graph is going to be shown.")

    if fig is None and ax is None:
        fig = plt.gcf()

    if ax is None:
        axes = fig.get_axes()
        if len(axes) != 0:
            if axes[0].name == '3d':
                fig = plt.figure()
                fig.add_subplot(111)
                ax = fig.get_axes()[0]
            else:
                ax = axes[0]
        else:
            fig.add_subplot(111)
            ax = fig.get_axes()[0]
    else:
        if ax.name == '3d':
            fig = plt.figure()
            fig.add_subplot(111)
            ax = fig.get_axes()[0]
        fig = ax.get_figure()

    return fig, ax


def plot_cluster_lines(estimator, X, fig=None, ax=None, sample_colors=None,
                       sample_labels=None, cluster_labels=None,
                       colormap=plt.cm.get_cmap('rainbow'), xlabel=None,
                       ylabel=None, title=None):
    """Implementation of the plotting of the results of the
    :func:`Fuzzy K-Means <fda.clustering.fuzzy_kmeans>` method.


    A kind of Parallel Coordinates plot is generated in this function with the
    membership values obtained from the algorithm. A line is plotted for each
    sample with the values for each cluster. See `Clustering Example
    <../auto_examples/plot_clustering.html>`_.

    Args:
        estimator (BaseEstimator object): estimator used to calculate the
            clusters.
        X (FDataGrd object): contains the samples which are grouped
            into different clusters.
        fig (figure object, optional): figure over which the graph is
            plotted in case ax is not specified. If None and ax is also None,
            the figure is initialized.
        ax (axis object, optional): axis over where the graph is  plotted.
            If None, see param fig.
        sample_colors (list of colors, optional): contains in order the colors of each
            sample of the fdatagrid.
        sample_labels (list of str, optional): contains in order the labels
            of each sample  of the fdatagrid.
        cluster_labels (list of str, optional): contains in order the names of each
            cluster the samples of the fdatagrid are classified into.
        colormap(colormap, optional): colormap from which the colors of the plot are taken.
        xlabel (str): Label for the x-axis. Defaults to "Sample".
        ylabel (str): Label for the y-axis. Defaults to "Membership grade".
        title (str, optional): Title for the figure where the clustering results are ploted.
            Defaults to "Membership grades of the samples to each cluster".

    Returns:
        (tuple): tuple containing:

            fig (figure object): figure object in which the graphs are plotted in case ax is None.

            ax (axes object): axes in which the graphs are plotted.

    """
    fdatagrid = X
    _check_if_estimator(estimator)

    if not isinstance(estimator, FuzzyKMeans):
        raise ValueError("The estimator must be a FuzzyKMeans object.")

    try:
        estimator._check_is_fitted()
        estimator._check_test_data(X)
    except NotFittedError:
        estimator.fit(X)

    fig, ax = _fig_and_ax_checks(fig, ax)

    _plot_clustering_checks(estimator, fdatagrid, sample_colors, sample_labels,
                            None, cluster_labels, None, None)

    xlabel, ylabel, title = _set_labels(xlabel, ylabel, title, "Cluster")

    if sample_colors is None:
        cluster_colors = colormap(np.arange(estimator.n_clusters) /
                                  (estimator.n_clusters - 1))
        labels_by_cluster = np.argmax(estimator.labels_, axis=1)
        sample_colors = cluster_colors[labels_by_cluster]

    if sample_labels is None:
        sample_labels = ['$SAMPLE: {}$'.format(i) for i in
                         range(fdatagrid.nsamples)]

    if cluster_labels is None:
        cluster_labels = ['${}$'.format(i) for i in
                          range(estimator.n_clusters)]


    ax.get_xaxis().set_major_locator(MaxNLocator(integer=True))
    for i in range(fdatagrid.nsamples):
        ax.plot(np.arange(estimator.n_clusters),
                   estimator.labels_[i],
                   label=sample_labels[i], color=sample_colors[i])
    ax.set_xticks(np.arange(estimator.n_clusters))
    ax.set_xticklabels(cluster_labels)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    datacursor(formatter='{label}'.format)

    fig.suptitle(title)
    return fig, ax


def plot_cluster_bars(estimator, X, fig=None, ax=None, sort=-1,
                      sample_labels=None, cluster_colors=None,
                      cluster_labels=None, colormap=plt.cm.get_cmap('rainbow'),
                      xlabel=None, ylabel=None, title=None):
    """Implementation of the plotting of the results of the
    :func:`Fuzzy K-Means <fda.clustering.fuzzy_kmeans>` method.


    A kind of barplot is generated in this function with the
    membership values obtained from the algorithm. There is a bar for each sample
    whose height is 1 (the sum of the membership values of a sample add to 1), and
    the part proportional to each cluster is coloured with the corresponding color.
    See `Clustering Example <../auto_examples/plot_clustering.html>`_.

    Args:
        estimator (BaseEstimator object): estimator used to calculate the
            clusters.
        X (FDataGrd object): contains the samples which are grouped
            into different clusters.
        fig (figure object, optional): figure over which the graph is
            plotted in case ax is not specified. If None and ax is also None,
            the figure is initialized.
        ax (axis object, optional): axis over where the graph is  plotted.
            If None, see param fig.
        sort(int, optional): Number in the range [-1, n_clusters) designating
            the cluster whose labels are sorted in a decrementing order.
            Defaults to -1, in this case, no sorting is done.
        sample_labels (list of str, optional): contains in order the labels
            of each sample  of the fdatagrid.
        cluster_labels (list of str, optional): contains in order the names of each
            cluster the samples of the fdatagrid are classified into.
        cluster_colors (list of colors): contains in order the colors of each
            cluster the samples of the fdatagrid are classified into.
        colormap(colormap, optional): colormap from which the colors of the plot are taken.
        xlabel (str): Label for the x-axis. Defaults to "Sample".
        ylabel (str): Label for the y-axis. Defaults to "Membership grade".
        title (str): Title for the figure where the clustering results are plotted.
            Defaults to "Membership grades of the samples to each cluster".

    Returns:
        (tuple): tuple containing:

            fig (figure object): figure object in which the graph is plotted in case ax is None.

            ax (axis object): axis in which the graph is plotted.

    """
    fdatagrid = X
    _check_if_estimator(estimator)

    if not isinstance(estimator, FuzzyKMeans):
        raise ValueError("The estimator must be a FuzzyKMeans object.")

    try:
        estimator._check_is_fitted()
        estimator._check_test_data(X)
    except NotFittedError:
        estimator.fit(X)

    if sort < -1 or sort >= estimator.n_clusters:
        raise ValueError(
            "The sorting number must belong to the interval [-1, n_clusters)")

    fig, ax = _fig_and_ax_checks(fig, ax)

    _plot_clustering_checks(estimator, fdatagrid, None, sample_labels,
                            cluster_colors, cluster_labels, None, None)

    xlabels, ylabels, title = _set_labels(xlabel, ylabel, title, "Sample")

    if sample_labels is None:
        sample_labels = np.arange(fdatagrid.nsamples)

    if cluster_colors is None:
        cluster_colors = colormap(
            np.arange(estimator.n_clusters) / (estimator.n_clusters - 1))

    if cluster_labels is None:
        cluster_labels = ['$CLUSTER: {}$'.format(i) for i in
                          range(estimator.n_clusters)]

    patches = []
    for i in range(estimator.n_clusters):
        patches.append(
            mpatches.Patch(color=cluster_colors[i], label=cluster_labels[i]))

    if sort != -1:
        sample_indices = np.argsort(-estimator.labels_[:, sort])
        sample_labels = np.copy(sample_labels[sample_indices])
        labels_dim = np.copy(estimator.labels_[sample_indices])


        temp_labels = np.copy(labels_dim[:, 0])
        labels_dim[:, 0] = labels_dim[:, sort]
        labels_dim[:, sort] = temp_labels

        temp_color = np.copy(cluster_colors[0])
        cluster_colors[0] = cluster_colors[sort]
        cluster_colors[sort] = temp_color
    else:
        labels_dim = estimator.labels_

    conc = np.zeros((fdatagrid.nsamples, 1))
    labels_dim = np.concatenate((conc, labels_dim), axis=-1)
    for i in range(estimator.n_clusters):
        ax.bar(np.arange(fdatagrid.nsamples),
                  labels_dim[:, i + 1],
                  bottom=np.sum(labels_dim[:, :(i + 1)], axis=1),
                  color=cluster_colors[i])
    ax.set_xticks(np.arange(fdatagrid.nsamples))
    ax.set_xticklabels(sample_labels)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.legend(handles=patches)

    fig.suptitle(title)
    return fig, ax



import unittest
import numpy as np

from skfda.representation.grid import FDataGrid
from skfda.ml.clustering.base_kmeans import KMeans, FuzzyKMeans


class TestClustering(unittest.TestCase):

    # def setUp(self): could be defined for set up before any test

    def test_kmeans_univariate(self):
        data_matrix = [[1, 1, 2, 3, 2.5, 2], [0.5, 0.5, 1, 2, 1.5, 1],
                       [-1, -1, -0.5, 1, 1, 0.5],
                       [-0.5, -0.5, -0.5, -1, -1, -1]]
        sample_points = [0, 2, 4, 6, 8, 10]
        fd = FDataGrid(data_matrix, sample_points)
        init = np.array([[0, 0, 0, 0, 0, 0], [2, 1, -1, 0.5, 0, -0.5]])
        init_fd = FDataGrid(init, sample_points)
        kmeans = KMeans(init=init_fd)
        kmeans.fit(fd)
        np.testing.assert_array_equal(kmeans.predict(fd),
                                      np.array([0, 0, 0, 1]))
        np.testing.assert_allclose(kmeans.transform(fd),
                                   np.array([[2.98142397, 9.23534876],
                                             [0.68718427, 6.50960828],
                                             [3.31243449, 4.39222798],
                                             [6.49679408, 0.]]))
        centers = FDataGrid(data_matrix=np.array(
            [[0.16666667, 0.16666667, 0.83333333, 2., 1.66666667, 1.16666667],
             [-0.5, -0.5, -0.5, -1., -1., -1.]]),
            sample_points=sample_points)
        np.testing.assert_array_almost_equal(
            kmeans.cluster_centers_.data_matrix,
            centers.data_matrix)
        np.testing.assert_allclose(kmeans.score(fd), np.array([-20.33333333]))
        np.testing.assert_array_equal(kmeans.n_iter_, np.array([3.]))

    # def test_kmeans_multivariate(self):
    #     data_matrix = [[[1, 0.3], [2, 0.4], [3, 0.5], [4, 0.6]],
    #                    [[2, 0.5], [3, 0.6], [4, 0.7], [5, 0.7]],
    #                    [[3, 0.2], [4, 0.3], [5, 0.4], [6, 0.5]]]
    #     sample_points = [2, 4, 6, 8]
    #     fd = FDataGrid(data_matrix, sample_points)
    #     kmeans = KMeans()
    #     kmeans.fit(fd)
    #     np.testing.assert_array_equal(kmeans.predict(fd),
    #                                   np.array([[1, 1],
    #                                             [1, 1],
    #                                             [0, 0]]))
    #     np.testing.assert_allclose(kmeans.transform(fd),
    #                                np.array([[[4.89897949, 0.24494897],
    #                                           [1.22474487, 0.23184046]],
    #                                          [[2.44948974, 0.70592729],
    #                                           [1.22474487, 0.23184046]],
    #                                          [[0., 0.],
    #                                           [3.67423461, 0.47478065]]]))
    #     centers = FDataGrid(data_matrix=np.array(
    #         [[[3, 0.2], [4, 0.3], [5, 0.4], [6, 0.5]],
    #          [[1.5, 0.4], [2.5, 0.5], [3.5, 0.6], [4.5, 0.65]]]),
    #         sample_points=sample_points)
    #     np.testing.assert_allclose(kmeans.cluster_centers_.data_matrix,
    #                                centers.data_matrix)
    #     np.testing.assert_allclose(kmeans.score(fd), np.array([-3., -0.1075]))
    #     np.testing.assert_array_equal(kmeans.n_iter_, np.array([2., 2.]))

    def test_fuzzy_kmeans_univariate(self):
        data_matrix = [[1, 1, 2, 3, 2.5, 2], [0.5, 0.5, 1, 2, 1.5, 1],
                       [-1, -1, -0.5, 1, 1, 0.5],
                       [-0.5, -0.5, -0.5, -1, -1, -1]]
        sample_points = [0, 2, 4, 6, 8, 10]
        fd = FDataGrid(data_matrix, sample_points)
        fuzzy_kmeans = FuzzyKMeans()
        fuzzy_kmeans.fit(fd)
        np.testing.assert_array_equal(fuzzy_kmeans.predict(fd),
                                      np.array([[0.965, 0.035],
                                                [0.94, 0.06],
                                                [0.227, 0.773],
                                                [0.049, 0.951]]))
        np.testing.assert_allclose(fuzzy_kmeans.transform(fd),
                                   np.array([[1.49228858, 7.87898791],
                                             [1.29380155, 5.12696975],
                                             [4.85542339, 2.63309793],
                                             [7.77455633, 1.75920889]]))
        centers = FDataGrid(data_matrix=np.array(
            [[0.7065078, 0.7065078, 1.45508111, 2.46698825,
              1.98143302, 1.48206743],
             [-0.69456401, -0.69456401, -0.49444239, -0.19713489,
              -0.19872214, -0.39844583]]), sample_points=sample_points)
        np.testing.assert_allclose(fuzzy_kmeans.cluster_centers_.data_matrix,
                                   centers.data_matrix)
        np.testing.assert_allclose(fuzzy_kmeans.score(fd),
                                   np.array([-13.928868250627902]))
        np.testing.assert_array_equal(fuzzy_kmeans.n_iter_, np.array([18.]))

    # def test_fuzzy_kmeans_multivariate(self):
    #     data_matrix = [[[1, 0.3], [2, 0.4], [3, 0.5], [4, 0.6]],
    #                    [[2, 0.5], [3, 0.6], [4, 0.7], [5, 0.7]],
    #                    [[3, 0.2], [4, 0.3], [5, 0.4], [6, 0.5]]]
    #     sample_points = [2, 4, 6, 8]
    #     fd = FDataGrid(data_matrix, sample_points)
    #     init = np.array([[[3, 0], [5, 0], [2, 0], [4, 0]],
    #                      [[0, 0], [0, 1], [0, 0], [0, 1]]])
    #     init_fd = FDataGrid(init, sample_points)
    #     fuzzy_kmeans = FuzzyKMeans(init=init_fd)
    #     fuzzy_kmeans.fit(fd)
    #     np.testing.assert_array_equal(fuzzy_kmeans.predict(fd),
    #                                   np.array([[[0., 1.],
    #                                              [0.5, 0.5]],
    #                                             [[1., 0.],
    #                                              [0.5, 0.5]],
    #                                             [[0.8, 0.2],
    #                                              [0.5, 0.5]]]))
    #     np.testing.assert_allclose(fuzzy_kmeans.transform(fd),
    #                                np.array([[[25., 1.26333333],
    #                                           [126.33333333, 1.26333333]],
    #                                          [[25., 2.45833333],
    #                                           [126.33333333, 2.45833333]],
    #                                          [[6., 0.78333333],
    #                                           [24., 0.78333333]]]))
    #     centers = FDataGrid(data_matrix=np.array(
    #         [[[2, 0], [3, 0], [4, 0], [5, 0]],
    #          [[1, 0], [2, 0], [3, 0], [4, 0]]]), sample_points=sample_points)
    #     np.testing.assert_allclose(fuzzy_kmeans.cluster_centers_.data_matrix,
    #                                centers.data_matrix)
    #     np.testing.assert_allclose(fuzzy_kmeans.score(fd), np.array(
    #         [-1.66211111e+04, -8.25302500e+00]))
    #     np.testing.assert_array_equal(fuzzy_kmeans.n_iter_,
    #                                   np.array([2., 2.]))


if __name__ == '__main__':
    print()
    unittest.main()



"""Magnitude-Shape Plot Module.

This module contains the necessary functions to construct the Magnitude-Shape Plot.
First the directional outlingness is calculated and then, an outliers detection method is implemented.

"""

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import scipy.integrate
from sklearn.covariance import MinCovDet
from scipy.stats import f, variation
from numpy import linalg as la
from io import BytesIO
import scipy

from ... import FDataGrid
from skfda.exploratory.depth import modified_band_depth

__author__ = "Amanda Hernando Bernabé"
__email__ = "amanda.hernando@estudiante.uam.es"


def directional_outlyingness(fdatagrid, depth_method=modified_band_depth,
                             dim_weights=None, pointwise_weights=None):
    r"""Computes the directional outlyingness of the functional data.

    Furthermore, it calculates both the mean and the variation of the
    directional outlyingness of the samples in the data set, which are also
    returned.

    The first one, the mean directional outlyingness, describes the relative
    position (including both distance and direction) of the samples on average
    to the center curve; its norm can be regarded as the magnitude outlyingness.

    The second one, the variation of the directional outlyingness, measures
    the change of the directional outlyingness in terms of both norm and
    direction across the whole design interval and can be regarded as the
    shape outlyingness.

    Firstly, the directional outlyingness is calculated as follows:

    .. math::
        \mathbf{O}\left(\mathbf{X}(t) , F_{\mathbf{X}(t)}\right) =
        \left\{\frac{1}{d\left(\mathbf{X}(t) , F_{\mathbf{X}(t)}\right)} - 1\right\} \cdot \mathbf{v}(t)

    where :math:`\mathbf{X}` is a stochastic process with probability distribution
    :math:`F`, :math:`d` a depth function and :math:`\mathbf{v}(t) = \left\{\mathbf{X}(t) -
    \mathbf{Z}(t)\right\} / \lVert \mathbf{X}(t) - \mathbf{Z}(t) \rVert`
    is the spatial sign of :math:`\left\{\mathbf{X}(t) - \mathbf{Z}(t)\right\}`,
    :math:`\mathbf{Z}(t)` denotes the median and ∥ · ∥ denotes the :math:`L_2` norm.

    From the above formula, we define the mean directional outlyingness as:

    .. math::
        \mathbf{MO}\left(\mathbf{X} , F_{\mathbf{X}}\right) = \int_I
        \mathbf{O}\left(\mathbf{X}(t) , F_{\mathbf{X}(t)}\right) \cdot w(t) dt ;

    and the variation of the directional outlyingness as:

    .. math::
        VO\left(\mathbf{X} , F_{\mathbf{X}}\right) = \int_I  \lVert\mathbf{O}\left(\mathbf{X}(t) ,
        F_{\mathbf{X}(t)}\right)-\mathbf{MO}\left(\mathbf{X} , F_{\mathbf{X}}\right)  \rVert^2 \cdot w(t) dt

    where :math:`w(t)` a weight function defined on the domain of :math:`\mathbf{X}`, :math:`I`.

    Then, the total functional outlyingness can be computed using these values:

    .. math::
        FO\left(\mathbf{X} , F_{\mathbf{X}}\right) = \lVert \mathbf{MO}\left(\mathbf{X} ,
        F_{\mathbf{X}}\right)\rVert^2 +  VO\left(\mathbf{X} , F_{\mathbf{X}}\right) .

    Args:
        fdatagrid (FDataGrid): Object containing the samples to be ordered according to
            the directional outlyingness.
        depth_method (:ref:`depth measure <depth-measures>`, optional): Method used to
            order the data. Defaults to :func:`modified band depth <fda.depth_measures.modified_band_depth>`.
        dim_weights (array_like, optional): an array containing the weights of each of
            the dimensions of the image. Defaults to the same weight for each of the
            dimensions: 1/ndim_image.
        pointwise_weights (array_like, optional): an array containing the weights of each
            point of discretisation where values have been recorded. Defaults to the same
            weight for each of the points: 1/len(interval).

    Returns:
        (tuple): tuple containing:

            dir_outlyingness (numpy.array((fdatagrid.shape))): List containing
            the values of the directional outlyingness of the FDataGrid object.

            mean_dir_outl (numpy.array((fdatagrid.nsamples, 2))): List containing
            the values of the magnitude outlyingness for each of the samples.

            variation_dir_outl (numpy.array((fdatagrid.nsamples,))): List
            containing the values of the shape outlyingness for each of the samples.

    Example:

        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2], [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5], [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> sample_points = [0, 2, 4, 6, 8, 10]
        >>> fd = FDataGrid(data_matrix, sample_points)
        >>> directional_outlyingness(fd)
        (array([[[ 1. ],
                [ 1. ],
                [ 1. ],
                [ 1. ],
                [ 1. ],
                [ 1. ]],
        <BLANKLINE>
               [[ 0. ],
                [ 0. ],
                [ 0. ],
                [ 0. ],
                [ 0. ],
                [ 0. ]],
        <BLANKLINE>
               [[-1. ],
                [-1. ],
                [-0.2],
                [-0.2],
                [-0.2],
                [-0.2]],
        <BLANKLINE>
               [[-0.2],
                [-0.2],
                [-0.2],
                [-1. ],
                [-1. ],
                [-1. ]]]), array([[ 1.66666667],
               [ 0.        ],
               [-0.73333333],
               [-1.        ]]), array([ 0.74074074,  0.        ,  0.36740741,  0.53333333]))


    """

    if fdatagrid.ndim_domain > 1:
        raise NotImplementedError("Only support 1 dimension on the domain.")

    if dim_weights is not None and (len(
            dim_weights) != fdatagrid.ndim_image or dim_weights.sum() != 1):
        raise ValueError(
            "There must be a weight in dim_weights for each dimension of the "
            "image and altogether must sum 1.")

    if pointwise_weights is not None and (len(
            pointwise_weights) != fdatagrid.ncol or pointwise_weights.sum() != 1):
        raise ValueError(
            "There must be a weight in pointwise_weights for each recorded time "
            "point and altogether must sum 1.")

    depth, depth_pointwise = depth_method(fdatagrid, pointwise=True)

    if dim_weights is None:
        dim_weights = np.ones(fdatagrid.ndim_image) / fdatagrid.ndim_image

    if pointwise_weights is None:
        pointwise_weights = np.ones(fdatagrid.ncol) / fdatagrid.ncol

    # Calculation of the depth of each multivariate sample with the
    # corresponding weight.
    weighted_depth = depth * dim_weights
    sample_depth = weighted_depth.sum(axis=-1)

    # Obtaining the median sample Z, to caculate v(t) = {X(t) − Z(t)}/∥ X(t) − Z(t)∥
    median_index = np.argmax(sample_depth)
    median = fdatagrid.data_matrix[median_index]
    v = fdatagrid.data_matrix - median
    v_norm = la.norm(v, axis=-1, keepdims=True)
    # To avoid ZeroDivisionError, the zeros are substituted by ones.
    v_norm[np.where(v_norm == 0)] = 1
    v_unitary = v / v_norm

    # Calculation of the depth of each point of each sample with
    # the corresponding weight.
    weighted_depth_pointwise = depth_pointwise * dim_weights
    sample_depth_pointwise = weighted_depth_pointwise.sum(axis=-1,
                                                          keepdims=True)

    # Calcuation directinal outlyingness
    dir_outlyingness = (1 / sample_depth_pointwise - 1) * v_unitary

    # Calcuation mean directinal outlyingness
    pointwise_weights_1 = np.tile(pointwise_weights,
                                  (fdatagrid.ndim_image, 1)).T
    weighted_dir_outlyingness = dir_outlyingness * pointwise_weights_1
    mean_dir_outl = scipy.integrate.simps(weighted_dir_outlyingness,
                                          fdatagrid.sample_points[0], axis = 1)

    # Calcuation variation directinal outlyingness
    mean_dir_outl_pointwise = np.repeat(mean_dir_outl, fdatagrid.ncol,
                                        axis=0).reshape(fdatagrid.shape)
    norm = np.square(
        la.norm(dir_outlyingness - mean_dir_outl_pointwise, axis=-1))
    weighted_norm = norm * pointwise_weights
    variation_dir_outl = scipy.integrate.simps(weighted_norm,
                                          fdatagrid.sample_points[0], axis=1)

    return dir_outlyingness, mean_dir_outl, variation_dir_outl

class MagnitudeShapePlot:
    r"""Implementation of the magnitude-shape plot

    This plot, which is based on the calculation of the
    :func:`directional outlyingness <fda.magnitude_shape_plot.directional_outlyingness>`
    of each of the samples, serves as a visualization tool for the centrality
    of curves. Furthermore, an outlier detection procedure is included.

    The norm of the mean of the directional outlyingness (:math:`\lVert\mathbf{MO}\rVert`)
    is plotted in the x-axis, and the variation of the directional outlyingness (:math:`VO`)
    in the y-axis.

    Considering :math:`\mathbf{Y} = \left(\mathbf{MO}^T, VO\right)^T`, the outlier detection method
    is implemented as described below.

    First, the square robust Mahalanobis distance is calculated based on a
    sample of size :math:`h \leq fdatagrid.nsamples`:

    .. math::
        {RMD}^2\left( \mathbf{Y}, \mathbf{\tilde{Y}}^*_J\right) = \left(
        \mathbf{Y} - \mathbf{\tilde{Y}}^*_J\right)^T  {\mathbf{S}^*_J}^{-1}
        \left( \mathbf{Y} - \mathbf{\tilde{Y}}^*_J\right)

    where :math:`J` denotes the group of :math:`h` samples that minimizes the
    determinant of the corresponding covariance matrix, :math:`\mathbf{\tilde{Y}}^*_J
    = h^{-1}\sum_{i\in{J}}\mathbf{Y}_i` and :math:`\mathbf{S}^*_J
    = h^{-1}\sum_{i\in{J}}\left( \mathbf{Y}_i - \mathbf{\tilde{Y}}^*_J\right)
    \left( \mathbf{Y}_i - \mathbf{\tilde{Y}}^*_J\right)^T`. The
    sub-sample of size h controls the robustness of the method.

    Then, the tail of this distance distribution is approximated as follows:

    .. math::
        \frac{c\left(m − p\right)}{m\left(p + 1\right)}RMD^2\left(
        \mathbf{Y}, \mathbf{\tilde{Y}}^*_J\right)\sim F_{p+1, m-p}

    where :math:`p` is the dmension of the image, and :math:`c` and :math:`m`
    are parameters determining the degrees of freedom of the :math:`F`-distribution
    and the scaling factor.

    .. math::
        c = E \left[s^*_{jj}\right]

    where :math:`s^*_{jj}` are the diagonal elements of MCD and

    .. math::
        m = \frac{2}{CV^2}

    where :math:`CV` is the estimated coefficient of variation of the diagonal elements of the  MCD shape estimator.

    Finally, we choose a cutoff value to determine the outliers, C ,
    as the α quantile of :math:`F_{p+1, m-p}`. We set :math:`\alpha = 0.993`,
    which is used in the classical boxplot for detecting outliers under a normal
    distribution.

    Attributes:
        fdatagrid (FDataGrid): Object to be visualized.
        depth_method (:ref:`depth measure <depth-measures>`, optional): Method
            used to order the data. Defaults to :func:`modified band depth
            <fda.depth_measures.modified_band_depth>`.
        dim_weights (array_like, optional): an array containing the weights
            of each of the dimensions of the image.
        pointwise_weights (array_like, optional): an array containing the
            weights of each points of discretisation where values have been
            recorded.
        alpha(float, optional): Denotes the quantile to choose the cutoff
            value for detecting outliers Defaults to 0.993, which is used
            in the classical boxplot.
        points(numpy.ndarray): 2-dimensional matrix where each row
            contains the points plotted in the graph.
        outliers (1-D array, (fdatagrid.nsamples,)): Contains 1 or 0 to denote
            if a sample is an outlier or not, respecively.
        colormap(matplotlib.pyplot.LinearSegmentedColormap, optional): Colormap
            from which the colors of the plot are extracted. Defaults to 'seismic'.
        color (float, optional): Tone of the colormap in which the nonoutlier
            points are  plotted. Defaults to 0.2.
        outliercol (float, optional): Tone of the colormap in which the outliers
            are plotted. Defaults to 0.8.
        xlabel (string, optional): Label of the x-axis. Defaults to 'MO',
            mean of the  directional outlyingness.
        ylabel (string, optional): Label of the y-axis. Defaults to 'VO',
            variation of the  directional outlyingness.
        title (string, optional): Title of the plot. defaults to 'MS-Plot'.

    Example:

        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2], [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5], [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> sample_points = [0, 2, 4, 6, 8, 10]
        >>> fd = FDataGrid(data_matrix, sample_points)
        >>> MagnitudeShapePlot(fd)
        MagnitudeShapePlot(
            FDataGrid=FDataGrid(
                array([[[ 1. ],
                        [ 1. ],
                        [ 2. ],
                        [ 3. ],
                        [ 2.5],
                        [ 2. ]],
        <BLANKLINE>
                       [[ 0.5],
                        [ 0.5],
                        [ 1. ],
                        [ 2. ],
                        [ 1.5],
                        [ 1. ]],
        <BLANKLINE>
                       [[-1. ],
                        [-1. ],
                        [-0.5],
                        [ 1. ],
                        [ 1. ],
                        [ 0.5]],
        <BLANKLINE>
                       [[-0.5],
                        [-0.5],
                        [-0.5],
                        [-1. ],
                        [-1. ],
                        [-1. ]]]),
                sample_points=[array([ 0,  2,  4,  6,  8, 10])],
                domain_range=array([[ 0, 10]]),
                dataset_label=None,
                axes_labels=None,
                extrapolation=None,
                interpolator=SplineInterpolator(interpolation_order=1, smoothness_parameter=0.0, monotone=False),
                keepdims=False),
            depth_method=modified_band_depth,
            dim_weights=None,
            pointwise_weights=None,
            alpha=0.993,
            points=array([[ 1.66666667,  0.74074074],
                   [ 0.        ,  0.        ],
                   [-0.73333333,  0.36740741],
                   [-1.        ,  0.53333333]]),
            outliers=array([0, 0, 0, 0]),
            colormap=seismic,
            color=0.2,
            outliercol=(0.8,),
            xlabel='MO',
            ylabel='VO',
            title='MS-Plot')
    """

    def __init__(self, fdatagrid, depth_method=modified_band_depth,
                 dim_weights=None, pointwise_weights=None, alpha=0.993,
                 assume_centered=False, support_fraction=None,random_state=0):
        """Initialization of the MagnitudeShapePlot class.

        Args:
            fdatagrid (FDataGrid): Object containing the data.
            depth_method (:ref:`depth measure <depth-measures>`, optional): Method
                used to order the data. Defaults to :func:`modified band depth
                <fda.depth_measures.modified_band_depth>`.
            dim_weights (array_like, optional): an array containing the weights
                of each of the dimensions of the image.
            pointwise_weights (array_like, optional): an array containing the
                weights of each points of discretisati on where values have been
                recorded.
            alpha (float, optional): Denotes the quantile to choose the cutoff
                value for detecting outliers Defaults to 0.993, which is used
                in the classical boxplot.
            assume_centered (boolean, optional): If True, the support of the robust
                location and the covariance estimates is computed, and a
                covariance estimate is recomputed from it, without centering
                the data. Useful to work with data whose mean is significantly
                equal to zero but is not exactly zero. If False, default value,
                the robust location and covariance are directly computed with
                the FastMCD algorithm without additional treatment.
            support_fraction (float, 0 < support_fraction < 1, optional): The proportion
                of points to be included in the support of the raw MCD estimate.
                Default is None, which implies that the minimum value of
                support_fraction will be used within the algorithm:
                [n_sample + n_features + 1] / 2
            random_state (int, RandomState instance or None, optional): If int,
                random_state is the seed used by the random number generator;
                If RandomState instance, random_state is the random number
                generator; If None, the random number generator is the
                RandomState instance used by np.random. By default, it is 0.

        """

        if fdatagrid.ndim_image > 1:
            raise NotImplementedError("Only support 1 dimension on the image.")

        # The depths of the samples are calculated giving them an ordering.
        _, mean_dir_outl, variation_dir_outl = directional_outlyingness(
            fdatagrid,
            depth_method,
            dim_weights,
            pointwise_weights)

        points = np.array(list(zip(mean_dir_outl, variation_dir_outl))).astype(
            float)

        # The square mahalanobis distances of the samples are calulated using MCD.
        cov = MinCovDet(store_precision=False, assume_centered=assume_centered,
                        support_fraction=support_fraction,
                        random_state=random_state).fit(points)
        rmd_2 = cov.mahalanobis(points)

        # Calculation of the degrees of freedom of the F-distribution
        # (approximation of the tail of the distance distribution).
        s_jj = np.diag(cov.covariance_)
        c = np.mean(s_jj)
        m = 2 / np.square(variation(s_jj))
        p = fdatagrid.ndim_image
        dfn = p + 1
        dfd = m - p

        # Calculation of the cutoff value and scaling factor to identify outliers.
        cutoff_value = f.ppf(alpha, dfn, dfd, loc=0, scale=1)
        scaling = c * dfd / m / dfn
        outliers = (scaling * rmd_2 > cutoff_value) * 1

        self._fdatagrid = fdatagrid
        self._depth_method = depth_method
        self._dim_weights = dim_weights
        self._pointwise_weights = pointwise_weights
        self._alpha = alpha
        self._mean_dir_outl = mean_dir_outl
        self._variation_dir_outl = variation_dir_outl
        self._points = points
        self._outliers = outliers
        self._colormap = plt.cm.get_cmap('seismic')
        self._color = 0.2
        self._outliercol = 0.8,
        self.xlabel = 'MO'
        self.ylabel = 'VO'
        self.title = 'MS-Plot'

    @property
    def fdatagrid(self):
        return self._fdatagrid

    @property
    def depth_method(self):
        return self._depth_method

    @property
    def dim_weights(self):
        return self._dim_weights

    @property
    def pointwise_weights(self):
        return self._pointwise_weights

    @property
    def alpha(self):
        return self._alpha

    @property
    def points(self):
        return self._points

    @property
    def outliers(self):
        return self._outliers

    @property
    def colormap(self):
        return self._colormap

    @colormap.setter
    def colormap(self, value):
        if not isinstance(value, matplotlib.colors.LinearSegmentedColormap):
            raise ValueError(
                "colormap must be of type matplotlib.colors.LinearSegmentedColormap")
        self._colormap = value

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        if value < 0 or value > 1:
            raise ValueError(
                "color must be a number between 0 and 1.")

        self._color = value

    @property
    def outliercol(self):
        return self._outliercol

    @outliercol.setter
    def outliercol(self, value):
        if value < 0 or value > 1:
            raise ValueError(
                "outcol must be a number between 0 and 1.")
        self._outliercol = value

    def plot(self, ax=None):
        """Visualization of the magnitude shape plot of the fdatagrid.

        Args:
            ax (axes object, optional): axes over where the graph is plotted.
                Defaults to matplotlib current axis.

        Returns:
            fig (figure object): figure object in which the graph is plotted.
            ax (axes object): axes in which the graph is plotted.

        """
        colors = np.zeros((self.fdatagrid.nsamples, 4))
        colors[np.where(self.outliers == 1)] = self.colormap(self.outliercol)
        colors[np.where(self.outliers == 0)] = self.colormap(self.color)

        if ax is None:
            ax = matplotlib.pyplot.gca()

        colors_rgba = [tuple(i) for i in colors]
        ax.scatter(self._mean_dir_outl, self._variation_dir_outl,
                   color=colors_rgba)

        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ylabel)
        ax.set_title(self.title)

        return ax.get_figure(), ax

    def __repr__(self):
        """Return repr(self)."""
        return (f"MagnitudeShapePlot("
                f"\nFDataGrid={repr(self.fdatagrid)},"
                f"\ndepth_method={self.depth_method.__name__},"
                f"\ndim_weights={repr(self.dim_weights)},"
                f"\npointwise_weights={repr(self.pointwise_weights)},"
                f"\nalpha={repr(self.alpha)},"
                f"\npoints={repr(self.points)},"
                f"\noutliers={repr(self.outliers)},"
                f"\ncolormap={self.colormap.name},"
                f"\ncolor={repr(self.color)},"
                f"\noutliercol={repr(self.outliercol)},"
                f"\nxlabel={repr(self.xlabel)},"
                f"\nylabel={repr(self.ylabel)},"
                f"\ntitle={repr(self.title)})").replace('\n', '\n    ')

    def _repr_svg_(self):
        plt.figure()
        fig, _ = self.plot()
        output = BytesIO()
        fig.savefig(output, format='svg')
        data = output.getvalue()
        plt.close(fig)
        return data.decode('utf-8')



import unittest

import numpy as np

from skfda.representation.basis import FDataBasis, Constant, Monomial
from skfda.misc import LinearDifferentialOperator


class TestBasis(unittest.TestCase):

    def test_init_integer(self):
        # Checks for a zero order Lfd object
        lfd_0 = LinearDifferentialOperator(order=0)
        weightfd = [FDataBasis(Constant((0, 1)), 1)]

        np.testing.assert_equal(lfd_0.order, 0,
                                "Wrong deriv order of the linear operator")
        np.testing.assert_equal(
            lfd_0.weights, weightfd,
            "Wrong list of weight functions of the linear operator")

        # Checks for a non zero order Lfd object
        lfd_3 = LinearDifferentialOperator(3)
        consfd = FDataBasis(Constant((0, 1)), np.identity(4)[3].reshape(-1, 1))
        bwtlist3 = consfd.to_list()

        np.testing.assert_equal(lfd_3.order, 3,
                                "Wrong deriv order of the linear operator")
        np.testing.assert_equal(
            lfd_3.weights, bwtlist3,
            "Wrong list of weight functions of the linear operator")

        np.testing.assert_raises(ValueError, LinearDifferentialOperator, -1)

    def test_init_list_int(self):
        coefficients = [1, 3, 4, 5, 6, 7]

        constant = Constant((0, 1))
        fd = FDataBasis(constant, np.array(coefficients).reshape(-1, 1))
        lfd = LinearDifferentialOperator(weights=coefficients)

        np.testing.assert_equal(lfd.order, 5,
                                "Wrong deriv order of the linear operator")
        np.testing.assert_equal(
            lfd.weights, fd.to_list(),
            "Wrong list of weight functions of the linear operator")

    def test_init_list_fdatabasis(self):
        weights = np.arange(4 * 5).reshape((5, 4))
        monomial = Monomial((0, 1), nbasis=4)
        fd = FDataBasis(monomial, weights)

        fdlist = [FDataBasis(monomial, weights[i])
                  for i in range(len(weights))]

        lfd = LinearDifferentialOperator(weights=fdlist)

        np.testing.assert_equal(lfd.order, 4,
                                "Wrong deriv order of the linear operator")
        np.testing.assert_equal(
            lfd.weights, fd.to_list(),
            "Wrong list of weight functions of the linear operator")

        contant = Constant((0, 2))
        fdlist.append(FDataBasis(contant, 1))
        np.testing.assert_raises(ValueError, LinearDifferentialOperator,
                                 None, fdlist)

    def test_init_wrong_params(self):
        np.testing.assert_raises(ValueError,
                                 LinearDifferentialOperator, 0, ['a'])
        np.testing.assert_raises(ValueError,
                                 LinearDifferentialOperator, 0, 'a')


if __name__ == '__main__':
    print()
    unittest.main()
